# -*- coding: utf-8 -*-
"""
Created on Sun Mar  5 13:49:24 2017

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import numpy as np
import warnings

from grapa.curve import Curve
from grapa.mathModule import roundSignificant


        
class CurveArrhenius(Curve):
    """
    Offers some support for fitting Arrhenius plots.
    A bit weid implementation: different types of data are handled with
    subclasses, but no instance is ever fabricated of these subclasses.
    Reason is, we don't knwo a priori which subclasse the instance will be,
    and we want to be able to easily switch from on type to the other.
    New types of plots can be implemented as subclasses without modifying the
    CurveArrhenius cain class.
    *** *** ***
    Generally assumes x as temperature [K] and y as a some kind of signal [eV].
    
    By default the output of the fit is E, Offset, with E the energy [eV] and
    Offset the value at 1/T = 0.
    The default fit function is Offset - E/(k*T), with k in [eV/K].
    
    Alternative implementations are:
     - _Arrhenius_variant is set to 'Cfdefects'
       Assumes input as omega [s-1] vs temperature [K].
       Useful to process C-f data. Relevant quantity to plot is
       log(omega * T**-2), and corresponding fit is log(2*ksi) - E /(kT).
    """
    
    CURVE = 'Curve Arrhenius'
    
    k_eVoverK = 8.6173303e-5 # [eV K-1]
    

    def __init__(self, data, attributes, silent=False):
        # main constructor
        Curve.__init__(self, data, attributes, silent=silent)
        self.update({'Curve': CurveArrhenius.CURVE})

    
    @classmethod
    def classNameGUI(cls):
        return cls.CURVE.replace('Curve', 'Fit')
    

    # GUI RELATED FUNCTIONS
    def funcListGUI(self, **kwargs):
        out = Curve.funcListGUI(self, **kwargs)
        # format: [func, 'func label', ['input 1', 'input 2', 'input 3', ...]]
        variant = self.getVariant()
        if self.getAttribute('_fitFunc', None) is None or self.getAttribute('_popt', None) is None:
            out.append([self.CurveArrhenius_fit, variant.BUTTONFitLabel(), [variant.BUTTONFitROI()], [[min(self.x_1000overK())*0.9, 1.1*max(self.x_1000overK())]]])
        else:
            lbl = variant.poptLabel
            param = roundSignificant(self.getAttribute('_popt'), 5)
            out.append([self.updateFitParam, 'Update fit', [lbl[0], lbl[1]], [param[0], '{:1.4e}'.format(param[1])]])
        # switch Arrhenius type
        subclasses = [variant.name, variant.switchModel, '', '== other possibilities ==']
        for child in CurveArrhenius.__subclasses__():
            if hasattr(child, 'name'):
                subclasses.append(child.name)
        out.append([self.setVariant, 'Switch model',
                    ['to model (now '+variant.name+')'],
                    [variant.switchModel], {},
                    [{'field':'Combobox', 'values':subclasses}]])
        # suggest default axis label
        if 'graph' in kwargs:
            alter = kwargs['graph'].getAttribute('alter')
            if alter in ['', ['','']]:
                if hasattr(variant, 'dataLabel'):
                    out.append([kwargs['graph'].updateValuesDictkeys,
                                'Save',
                                ['xlabel', 'ylabel'],
                                variant.dataLabel,
                                {'keys': ['xlabel', 'ylabel']}])
            elif alter == ['CurveArrhenius.x_1000overK', 'CurveArrhenius.y_Arrhenius']:
                if hasattr(variant, 'dataLabelArrhenius'):
                    out.append([kwargs['graph'].updateValuesDictkeys,
                                'Save',
                                ['xlabel', 'ylabel'],
                                variant.dataLabelArrhenius,
                                {'keys': ['xlabel', 'ylabel']}])
                
        out.append([self.printHelp, 'Help!', [], []])
        return out
    
    def alterListGUI(self):
        variant = self.getVariant()
        if hasattr(variant, 'alterListGUIspecific'):
            return getattr(variant, 'alterListGUIspecific')(self)
        out = Curve.alterListGUI(self)
        out += [['Arrhenius vs 1000/T', ['CurveArrhenius.x_1000overK', 'CurveArrhenius.y_Arrhenius'], '']]
        return out
    
    # these one can be overloaded
    @classmethod
    def BUTTONFitLabel(cls):
        return 'Fit Arrhenius'
    @classmethod
    def BUTTONFitROI(cls):
        return 'ROI [1000/K]'
    
    def x_1000overK(self, index=np.nan, **kwargs):
        """ x-axis transform function """
        variant = CurveArrhenius.getVariant(self) # syntax: not necessarily a CurveArrhenius
        if variant != CurveArrhenius and 'x_1000overK' in variant.__dict__: # maybe to do a non-Arrhenius fit?
            return variant.x_1000overK(self, index=index, **kwargs)
        return 1000 / self.x(index, **kwargs)
    
    def y_Arrhenius(self, index=np.nan, **kwargs):
        """ y-axis transform function, getting the Arrhenius straight line """
        variant = CurveArrhenius.getVariant(self) # syntax: not necessarily a CurveArrhenius
        if variant != CurveArrhenius and 'y_Arrhenius' in variant.__dict__: # only if y_Arrhenius explicitely defined, not inherited
            return variant.y_Arrhenius(self, index=index, **kwargs)
        print('CurveArrhenius y_Arrhenius, normally should not come to this point!')
        return self.y(index, **kwargs)
    
    # handling of subclasses of CurveArrhenius
    def getVariant(self):
        var = self.getAttribute('_Arrhenius_variant')
        subclasses = CurveArrhenius.__subclasses__()
        for i in range(len(subclasses)):
            if var == subclasses[i].name:
                return subclasses[i]
        return CurveArrheniusDefault
        
    def setVariant(self, newVariant):
        subclasses = CurveArrhenius.__subclasses__()
        for i in range(len(subclasses)):
            if newVariant == subclasses[i].name:
                self.update({'_Arrhenius_variant': newVariant})
                return True
        if newVariant not in ['', 'CurveArrhenius', 'default']:
            print('ERROR Arrhenius: cannot find model', newVariant,', going for default.')
        self.update({'_Arrhenius_variant': 'default'})
        return True
        
    # functions for Arrhenius plot
    def CurveArrhenius_fit(self, Tlim=None, silent=False):
        """ Returns a Curve based on a fit on Arrhenius plot. """
        variant = self.getVariant()
        popt = self.fit_Arrhenius(Tlim=Tlim, silent=silent)
        attr = {'color': 'k', '_popt': popt, '_fitFunc': 'func_Arrhenius',
                '_Arrhenius_variant': variant.name}
        lbl = variant.poptLabel
        if not silent:
            print('Fit results:', lbl[0],'=', lbl[2].format(popt[0]) + ',',
                                  lbl[1],'=', lbl[3].format(popt[1]) + '.')
        return CurveArrhenius([self.x(), self.func_Arrhenius(self.x(), *popt)], attr)
        
    def _fit_Arrhenius_getData(self, Tlim=None):
        T = self.x_1000overK()
        if Tlim is None:
            Tlim = [min(T), max(T)]
        mask = np.ones(len(self.x()), dtype=bool)
        for i in range(len(mask)):
            if T[i] < Tlim[0] or T[i] > Tlim[1]:
                mask[i] = False
        datax = self.x_1000overK()[mask]
        datay = self.y_Arrhenius()[mask]
        return datax, datay
    def fit_Arrhenius(self, Tlim=None, silent=False):
        """
        Fits the data on the Arrhenius plot, in ROI Tlim[0] to Tlim[1].
        Returns Energy barrier [eV] and prefactor.
        """
        datax, datay = self._fit_Arrhenius_getData(Tlim=Tlim)
        z = np.polyfit(datax, datay, 1, full=True)[0]
        # output
        variant = self.getVariant()
        if not silent:
            print('CurveArrhenius fit variant', variant.name,', assumed input as',
                  variant.dataLabel[1], 'vs', variant.dataLabel[0])
            print(variant.getHelp())
        return variant.popt(z) # [-z[0]*CurveArrhenius.k_eVoverK * 1000, z[1]]
    
    def func_Arrhenius(self, T, E, prefactor):
        """
        Returns function(K) which will appear linear on a Arrhenius plot.
        E is energy barrier
        prefactor is value on plot at 1/T = 0
        """
        variant = self.getVariant()
        if hasattr(variant, 'func_Arrhenius'):
            return variant.func_Arrhenius(T, E, prefactor)
        print('CurveArrhenius func_Arrhenius, normally should not come to this point!')
        return prefactor - E / (CurveArrhenius.k_eVoverK * T)
            

    def printHelp(self):
        print('*** *** ***')
        print('Curve Arrhenius general help:')
        print('Offers some support for fitting Arrhenius plots.')
        print('Generally assumes x as temperature [K] and y as a some kind of signal [eV].')
        print('By default the output of the fit is E, Offset, with E the energy [eV] and Offset the value at 1/T = 0.')
        print('The default fit function is Offset - E/(k*T), with k in [eV/K].')
        print('')
        print('List of alternative implementations:')
        dct = {}
        variant = self.getVariant()
        subclasses = CurveArrhenius.__subclasses__()
        for cls in subclasses:
            if cls != variant:
                dct.update({cls.name: cls})
        lst = [key for key in dct]
        lst.sort()
        for key in lst:
            print(key, ':', dct[key].getHelp(short=True))
        print('')
        print('Help on this specific implementation of Arrhenius Curve:')
        print(variant.getHelp())
        dataLabel = self.getVariant().dataLabel
        print('Expected data input:', dataLabel[0], ',', dataLabel[1])
        return True


        
        
class CurveArrheniusDefault(CurveArrhenius):
    """
    The instances will be of CurveArrhenius, but main class will call the
    methods implemented in the desired child classes.
    """
    name = 'default' # identifier
    dataLabel = ['Temperature [K]', 'Value [eV]']
    poptLabel = ['E [eV]', 'Offset [eV]', '{:1.4e}', '{:.4f}']
    switchModel = '' # sggestion to user, a other model to try on his data
    # default Curve attributes such that the CurveArrhenius is correctly defined
    attr = {'Curve': 'Curve Arrhenius', '_Arrhenius_variant': 'default',
            'label': 'value vs Temperature'}
    def popt(z): # transforms polynom z into correct fit parameters
        return [-z[0]*CurveArrhenius.k_eVoverK * 1000, z[1]]
    def getHelp(short=False): # the help dedicated to the specific implementation
        return 'Arrhenius plot is: value = offset - E /(kT)'
    def y_Arrhenius(self, index=np.nan, **kwargs):
         return self.y(index, **kwargs)
    def func_Arrhenius(T, E, prefactor):
        return prefactor - E / (CurveArrhenius.k_eVoverK * T)
    
class CurveArrheniusJscVocJ00(CurveArrhenius):
    name = 'JscVocJ00'
    dataLabel = ['A(T) * Temperature [K]', 'J0 [mA cm-2]']
    poptLabel = ['E_A [eV]', 'J00 [mA cm-2]', '{:1.4f}', '{:1.4f}']
    switchModel = 'JscVocJ00'
    attr = {'Curve': 'Curve Arrhenius', '_Arrhenius_variant': 'JscVocJ00',
            'label': 'J0 [mA cm-2] vs A(t)*T [K]',
            'linestyle': 'none', 'linespec': 'o'}
    def popt(z): # fit polynom z to desired output values
        return [-z[0]*CurveArrhenius.k_eVoverK * 1000, np.exp(z[1])]
    def func_Arrhenius(T, E, prefactor): # replace default fit function
        return np.exp(np.log(prefactor) - E / (CurveArrhenius.k_eVoverK * T))
    def y_Arrhenius(self, index=np.nan, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            out = np.log(self.y(index=index, **kwargs))
        return out
    def getHelp(short=False):
        out = 'Computes E_A the activation energy of J0, and J00 its value extrapolated at 1/T = 0.'
        if not short:
            out+= '\nThe equation is: log(J0) = log(J00) - E_A / (k AT),\n'
            out+= '   stemming from J0 = J00 * exp(- E_A / (k AT))'
        return out


class CurveArrheniusCfDefects(CurveArrhenius):
    """ 'fit': 'ln2ksi-E/kT', 'y': 'ln(omega*T**-2)' """
    name = 'Cfdefects'
    dataLabel = ['Temperature [K]', 'omega [s-1]']
    poptLabel = ['E [eV]', 'ksi', '{:1.5f}', '{:.4e}']
    switchModel = 'Cfdefault'
    attr = {'Curve': 'Curve Arrhenius', '_Arrhenius_variant': 'Cfdefects',
            'label': 'omega vs Temperature', 'linestyle': 'none'}
            # by default hidden - would cause strange value on derivative plot
    dataLabelArrhenius = [r'1000 / T', r'ln($\omega T^{-2}$)']
    def popt(z):
        return [-z[0] * CurveArrhenius.k_eVoverK * 1000, 0.5 * np.exp(z[1])]
    def func_Arrhenius(T, E, prefactor):
        return T**2 * np.exp(np.log(2 * prefactor) - E / (CurveArrhenius.k_eVoverK * T))
    def y_Arrhenius(self, index=np.nan, **kwargs):
        # must compute omega T**-2 and show log, corresponding fit expression is 'log2ksi-omega/kT'
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            out = np.log(self.y(index, **kwargs) * self.x(index, **kwargs)**-2)
        return out
    def getHelp(short=False):
        if short:
            return 'Handles omega [s-1] vs T [K] inflection points of C-f curves, using a weak T**-2 temperature dependency.'
        out = 'Assumes traps can follow frequency: omega = 2 ksi T**2 exp(-E/(kT))\n'
        out+= '    then Arrhenius plot is: ln(omega * T**-2) = ln(2*ksi) - E /(kT)'
        return out

     
class CurveArrheniusCfdefault(CurveArrhenius):
    """ 'fit': 'ln2ni-E/kT', 'y': 'ln(omega)' """
    name = 'Cfdefault'
    dataLabel = ['Temperature [K]', 'omega [s-1]']
    poptLabel = ['E [eV]', 'nu', '{:1.5f}', '{:.4e}']
    switchModel = 'Cfdefects'
    attr = {'Curve': 'Curve Arrhenius', '_Arrhenius_variant': 'Cfdefault',
            'label': 'omega vs Temperature', 'linestyle': 'none'}
            # by default hidden - would cause strange value on derivative plot
    dataLabelArrhenius = [r'1000 / T', 'ln($\omega$)']
    def popt(z):
        return [-z[0] * CurveArrhenius.k_eVoverK * 1000, 0.5 * np.exp(z[1])]
    def func_Arrhenius(T, E, prefactor):
        return np.exp(np.log(2 * prefactor) - E / (CurveArrhenius.k_eVoverK * T))
    def y_Arrhenius(self, index=np.nan, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            out = np.log(self.y(index, **kwargs))
        return out
    def getHelp(short=False):
        if short:
            return 'Handles omega [s-1] vs T [K] inflection points of C-f curves, using no weak T**-2 temperature dependency.'
        out = 'Assumes a (barrier / trap / something without weak T dependency) can follow frequency:\n'
        out+= 'omega = 2 nu exp(-E/(kT))\n'
        out+='    then Arrhenius plot is: ln(omega) = ln(2*nu) - E /(kT)'
        return out


class CurveArrheniusExtrapolToZero(CurveArrhenius):
    """
    This is not an Arrhenius, but we can recyle the same code for any 1-degree
    polynomial interpolation.
    """
    @classmethod
    def BUTTONFitLabel(cls):
        return 'Fit extrapolate to 0'
    @classmethod
    def BUTTONFitROI(cls):
        return 'ROI [K]'
    name = 'ExtrapolationTo0' # is NOT an Arrhenius fit, still, can use polynom fit
    dataLabel = ['Temperature [K]', 'Value [unit]']
    poptLabel = ['Increase [unit / K]', 'Value at 0 [unit]', '{:.4e}', '{:.4e}']
    switchModel = ''
    attr = {'Curve': 'Curve Arrhenius', '_Arrhenius_variant': 'ExtrapolationTo0'}
    dataLabelArrhenius = ['Temperature [K]', 'Value [unit]']
    def popt(z):
        return [z[0], z[1]]
    def func_Arrhenius(T, E, prefactor):
        return prefactor + E * T
    def y_Arrhenius(self, index=np.nan, **kwargs):
        return self.y(index, **kwargs)
    def x_1000overK(self, index=np.nan, **kwargs): # NOT ARRHENIUS
        return self.x(index, **kwargs)
    def getHelp(short=False):
        out = '(not Arrhenius relation) Fits the data with 1-degree polynom.'
        if short:
            return out
        out += '\nThe equation is: value = ValueAt0 + Increase * T'
        return out
    

class CurveArrheniusExpDecay(CurveArrhenius):
    """
    This is not an Arrhenius, but we can recyle the same code for another
    expression
    Created to handle Urbach decays of EQE curves 
    """
    @classmethod
    def BUTTONFitLabel(cls):
        return 'Fit exp. decay'
    @classmethod
    def BUTTONFitROI(cls):
        return 'ROI [eV]'
    name = 'expDecay' # is NOT an Arrhenius fit, still, can use polynom fit
    dataLabel = ['Wavelength [nm]', 'Value [unit]']
    poptLabel = ['Decay energy [meV]', 'Energy @ 100% [eV]', '{:.4e}', '{:.4e}']
    switchModel = ''
    attr = {'Curve': 'Curve Arrhenius', '_Arrhenius_variant': 'expDecay'}
    dataLabelArrhenius = ['Energy [eV]', 'log(value)']
    def alterListGUIspecific(self):
        out = Curve.alterListGUI(self)
        return out
    def popt(z):
        return [1/z[0]*1000, -z[1]/z[0]]
    def func_Arrhenius(nm, U, E_at_1):
        return np.exp(((Curve.NMTOEV / nm) - (E_at_1)) / (U*0.001))
    def y_Arrhenius(self, index=np.nan, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            return np.log(self.y(index, **kwargs))
    def x_1000overK(self, index=np.nan, **kwargs): # NOT ARRHENIUS
        return self.x(index, alter='nmeV', **kwargs)
    def getHelp(short=False):
        out = 'Characterize an exponential decay (ie. Urbach).'
        if short:
            return out
        out += '\nThe equation is: value = exp((E - E_at_1) / (E_decay/1000))'
        return out
    



# -*- coding: utf-8 -*-
"""
Created on Fri Jul 15 15:46:13 2016

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""
import os
import numpy as np
from copy import deepcopy
from re import findall as refindall
import matplotlib as mpl

from grapa.curve import Curve
from grapa.mathModule import is_number, stringToVariable


class Graph:
    """
## TO UPDATE
     content of the class
    'filename': str
    'data': list of Curve
    'headers': dict containing various information about the data
    'graphInfo': dict containing information related to graph.
                 Example: xlim, xlabel, etc.
    'sampleInfo': Deprecated. Dict containing various information about the
                  sample. Generally is empty, as the relevant data are stored
                  in the Curve object.
    
    List of useful methods
    __init__(self, filename, complement='', silent=False)
    __str__(self)

    length(self)
    merge(self, graph)
    append(self, curve)
	
    curve(self, index)
    curves(self, attr, value, strLower=False, strStartWith=False):
    iterCurves(self)
    deleteCurve(self, i)
    replaceCurve(self, newCurve, idx)
    swapCurves(self, idx1, idx2, relative=False)
    duplicateCurve(self, idx1)
    castCurve(self, newType, idx, silentSuccess=False)

    update(self, attributes, ifAll=False)
    getAttribute(self, attr, default='')
	replaceLabels(self, old, new)
    colorize(self, colorscale, sameIfEmptyLabel=False, avoidWhite=False)
    applyTemplate(self, graph)
    
    formatAxisLabel(self, default)
    
    config(self, key, default='', astype='auto')

	
    export(self, filesave='', saveAltered=False, ifTemplate=False, ifCompact=True, ifClipboardExport=False)
    plot(self, filesave='', imgFormat='', figsize=(0, 0), ifSave=True, ifExport=True, figAx=None, ifSubPlot=False)
    """
    # some class constants
    FIGSIZE_DEFAULT = (6.0, 4.0)
    FILEIO_DPI = 80
    DEFAULT = {}
    DEFAULT['subplots_adjust'] = 0.15
    DEFAULT['fontsize'] = mpl.rcParams['font.size']
    # Default axis labels. Normally not used, but can be overidden and used in
    # subclasses. Possible data formats are:
    # ['x axis label [unit]', 'y axis label [unit]']
    # [['Quantity x', 'Q', 'unit'], ['Quantity y', 'Q', 'unit']]. Q will be placed in between $ $
    AXISLABELS = ['', ''] 
    
    
    headersKeys = ['meastype', 'sample', 'collabels'] + ['savesilent']
    graphInfoKeysData = []
    graphInfoKeysData.append(['== Figure ==', ''])
    graphInfoKeysData.append(['figsize',         'Figure size (inch).\nExample: "(6.0, 4.0)"'])
    graphInfoKeysData.append(['subplots_adjust', 'Margins (relative).\nExamples: "0.15" (bottom only), or "[0.125, 0.1, 0.9, 0.9]" left,b,r,top, or "[1,1,5,3.5,\'abs\']"'])
    graphInfoKeysData.append(['dpi',             'Resolution dots-per-inch.\nExample: "300"'])
    graphInfoKeysData.append(['fontsize',        'Font size of titles, annotations, etc.\nExample: "12"'])
    graphInfoKeysData.append(['title',           'Graph title, based on ax.set_title().\nExample: "My data", or "[\'A dataset\', {\'color\':\'r\'}]"'])
    graphInfoKeysData.append(['== Axes ==', ''])
    graphInfoKeysData.append(['xlim',            'Limits of x axis, based on ax.set_xlim().\nExamples: "[2,9]", or "[\'\',4]"'])
    graphInfoKeysData.append(['ylim',            'Limits of y axis, based on ax.set_ylim().\nExamples: "[0,100]", or "[0,\'\']"'])
    graphInfoKeysData.append(['xlabel',          'Label of x axis, based on ax.set_xlabel().\nExample: "Axis x [unit]", "[\'My label\', {\'size\':6, \'color\':\'r\'}]"'])
    graphInfoKeysData.append(['ylabel',          'Label of y axis, based on ax.set_ylabel().\nExample: "Axis y [unit]", "[\'My label\', {\'size\':6, \'color\':\'r\'}]"'])
    graphInfoKeysData.append(['xticksstep',      'Value difference between ticks on x axis, or ticks positions.\nExample: "0.01", or "[0,1,2]"']) # ax.xaxis.set_ticks
    graphInfoKeysData.append(['yticksstep',      'Value difference between ticks on y axis, or ticks positions.\nExample: "0.01", or "[0,1,2]"']) # ax.yaxis.set_ticks
    graphInfoKeysData.append(['xtickslabels',    'Customized ticks. First is a list of values, then a list of labels, then possibly options.\nExamples: "[[0,1],[\'some\',\'value\']]", or "[None, None, {\'rotation\':45, \'size\': 6, \'color\':\'r\'}]"']) # plt.xticks
    graphInfoKeysData.append(['ytickslabels',    'Customized ticks. First is a list of values, then a list of labels, then possibly options.\nExamples: "[[0,1],[\'some\',\'value\']]", or "[None, None, {\'rotation\':45, \'size\': 6, \'color\':\'r\'}]"']) # plt.yticks
    graphInfoKeysData.append(['xlabel_coords',   'Position of xlabel, based on ax.xaxis.set_label_coords().\nExamples: "-0.1", or "[0.5,-0.15]"']) 
    graphInfoKeysData.append(['ylabel_coords',   'Position of ylabel, based on ax.yaxis.set_label_coords().\nExamples: "-0.1", or "[-0.1,0.5]"'])
    graphInfoKeysData.append(['== Legends ==', ''])
    graphInfoKeysData.append(['legendproperties','Position, or keywords to ax.legend(). Examples: "best", "sw", or\n"{\'bbox_to_anchor\':(0.2,0.8), \'ncol\':2, \'fontsize\':8}"'])
    graphInfoKeysData.append(['legendtitle',     'Legend title. Example: "Some title", or "[\'Some title\', {\'size\':25}]"']) # messy implementation
    graphInfoKeysData.append(['== Annotations ==', ''])
    graphInfoKeysData.append(['axhline',         'Horizontal line(s), based on ax.axhline().\nExample: "0", or "[-1, 1]"'])
    graphInfoKeysData.append(['axvline',         'Vertical line(s), based on ax.axvline().\nExample: "0", or "[-1, 1]"'])
    graphInfoKeysData.append(['text',            'Annotations, use GUI window if possible. "Some text", or "[\'Here\', \'There\']"'])
    graphInfoKeysData.append(['textxy',          'Use GUI window if possible. "(0.05, 0.95)" , or "[(0.2, 0.3), (0.8, 0.9)]"'])
    graphInfoKeysData.append(['textargs',        'Use GUI window if possible. "{\'fontsize\':15}", or "[{\'horizontalalignment\': \'right\',\n\'xytext\': (0.4, 0.65), \'arrowprops\': {\'shrink\': 0.05}, \'xy\': (0.46, 0.32)}, {}]"'])
    graphInfoKeysData.append(['== Secondary axes ==', ''])
    graphInfoKeysData.append(['twiny_xlabel',    'Secondary x axis label.\nExamples: "Other axis x [unit]", "[\'My label\', {\'size\':6, \'color\':\'r\'}]"'])
    graphInfoKeysData.append(['twinx_ylabel',    'Secondary y axis label.\nExamples: "Other axis x [unit]", "[\'My label\', {\'size\':6, \'color\':\'r\'}]"'])
    graphInfoKeysData.append(['twiny_xlim',      'Secondary x axis limits.\nExamples: "[2,9]", or "[\'\',4]"'])
    graphInfoKeysData.append(['twinx_ylim',      'Secondary y axis limits.\nExamples: "[0,100]", or "[0,\'\']"'])
    graphInfoKeysData.append(['== Misc ==', ''])
    graphInfoKeysData.append(['alter',           'Data transform keyword, specific to the type of manipulated data.\nExamples: "linear", or "[\'nmeV\', \'tauc\']"'])
    graphInfoKeysData.append(['typeplot',        'General graph plotting instruction, based on ax.set_xscale() and ax.set_yscale().\nExamples: "plot", "semilogx", etc.'])
    graphInfoKeysData.append(['arbitraryfunctions', 'A list of instructions. Each instruction is a list as\n[method of ax, list of arguments, dict of keyword arguments]', ["[[\'xaxis.set_ticks\',[[1.5, 5.5]],{\'minor\':True}], [\'set_xticklabels\',[[\'a\',\'b\']],{\'minor\':True}]]", "[['set_axis_off', [], {}]]", "[['yaxis.set_major_formatter', ['StrMethodFormatter({x:.2f})'], {}]]", "[['xaxis.set_minor_locator', ['MultipleLocator(0.5)'], {}]]"]])
    graphInfoKeysData_ = np.array([d[0:2] for d in graphInfoKeysData])
    graphInfoKeys = list(graphInfoKeysData_[:,0])
    graphInfoKeysExample = list(graphInfoKeysData_[:,1])
    graphInfoKeysExalist = []
    for i in range(len(graphInfoKeysData)):
        if len(graphInfoKeysData[i]) < 3:
            split = graphInfoKeysData[i][1].split('"')
            graphInfoKeysExalist.append(['']+split[1::2])
        else:
            graphInfoKeysExalist.append(graphInfoKeysData[i][2])
    
#    graphInfoKeys = ['xlim', 'ylim', 'figsize', 'title', 'xlabel', 'ylabel'] + ['text', 'textxy', 'textargs'] + ['xticksstep', 'yticksstep'] + ['axhline', 'axvline'] + ['legendproperties', 'legendtitle'] + ['alter', 'typeplot'] + ['twinx_ylabel', 'twinx_ylim', 'twiny_xlabel', 'twiny_xlim'] + ['xlabel_coords', 'ylabel_coords'] + ['subplots_adjust', 'fontsize'] + ['xtickslabels', 'ytickslabels'] + ['arbitraryfunctions'] + ['dpi']
#    graphInfoKeysExample = ["[2,9], or ['',4]", "[0,100], or [0,'']", str(FIGSIZE_DEFAULT)+' (inch)', 'Some title', 'Axis x [unit]', 'Axis y [unit]'] + ['Some text, or [\'Here\', \'There\']', "(0.05, 0.95) , or [(0.2, 0.3), (0.8, 0.9)] if multiple text", "{'fontsize':15}, or \n[{'horizontalalignment': 'right', 'xytext': (0.4, 0.65),\n'arrowprops': {'shrink': 0.05}, 'xy': (0.46, 0.32)}, {}]"] + ['0.01, or [0,1,2]', '0.01, or [0,1,2]'] + ['0, or [-1, 1]', '0, or [-1, 1]'] + ["best, sw, etc., or\n{'bbox_to_anchor':(0.2,0.8),'ncol':2, 'fontsize':8} kwargs to ax.legend()", 'Some title, or [\'Some title\', {\'size\':25}] font properties prop'] + ["linear, or ['nmeV', 'tauc']", 'plot, fill, scatter, boxplot, etc.'] + ['Secondary y axis label [unit]', "[0,100], or [0,'']", 'Secondary x axis label [unit]', "[2,9], or ['',4]"] + ['-0.15 or [0.5,-0.15]', '-0.1 or [-0.1,0.5]'] + ['0.15 (bottom only), or [0.125, 0.1, 0.9, 0.9] left,b,r,top, or [1,1,5,3.5,\'abs\']', '12'] + ["[[0,1],['some','value']], or [None, None, {'rotation':45, 'color':'r'}]", "[[0,1],['some','value']], or [None, None, {'rotation':45, 'color':'r'}]"] + ["[['xaxis.set_ticks',[[1.5, 5.5]],{'minor':True}],\n['set_xticklabels',[['a','b']],{'minor':True}]]"] + [300]
    type_examples = ['', '== usual methods ==']
    # "simple" plotting methods, with prototype similar to plot()
    type_examples += ['plot', 'fill', 'errorbar', 'scatter', 'boxplot']
    type_examples += ['== similar to plot ==', 'semilogx', 'semilogy', 'loglog', 'plot_date', 'stem', 'step', 'triplot']
    # plotting methods not accepting formatting string as 3rd argument
    type_examples += ['== (no linespec) ==', 'bar', 'barbs', 'barh', 'cohere', 'csd', 'fill_between', 'fill_betweenx', 'hexbin', 'hist2d', 'quiver', 'xcorr']
    #  plotting of single vector data
    type_examples += ['== 1D vector data ==', 'acorr', 'angle_spectrum', 'eventplot', 'hist', 'magnitude_spectrum', 'phase_spectrum', 'pie', 'psd', 'specgram']
    type_examples += ['== other ==', 'spy', 'stackplot', 'violinplot']
    type_examples += ['imshow', 'contour', 'contourf']
    # list of attributes recognized for plotting. These are typically attributes of the curves. Can be extended if needed.
    dataInfoKeysGraphData = []
    dataInfoKeysGraphData.append(['== Basic properties ==',  ''])
    dataInfoKeysGraphData.append(['type',           'Plotting method of Axes, ie. "plot", "scatter", "fill", "boxplot", "errorbar", etc.\nTip: after "scatter", set next Curve as "scatter_c" or "scatter_s"', type_examples])
    dataInfoKeysGraphData.append(['linespec',       'Format string controlling the line style or marker.\nExamples: "r", "--", ".-or", "^k"'])
    dataInfoKeysGraphData.append(['label',          'Curve label to be shown in legend. Example: "Experiment C"'])
    dataInfoKeysGraphData.append(['== Display ==',  ''])
    dataInfoKeysGraphData.append(['color',          'Curve color. Examples: "r", "purple", or "[0.5,0,0]" (rgb or rgba notations)'])
    dataInfoKeysGraphData.append(['alpha',          'Transparency. Examples. "0.5", "1"'])
    dataInfoKeysGraphData.append(['linewidth',      'Linewidth in points. Example: "1.5"'])
    dataInfoKeysGraphData.append(['marker',         'Examples: "o", "s", "x"'])
    dataInfoKeysGraphData.append(['markersize',     'Size of marker, in points. Example: "2.5"'])
    dataInfoKeysGraphData.append(['markerfacecolor','Marker inner color. Example: "r", "[0.5,0,0]"'])
    dataInfoKeysGraphData.append(['markeredgecolor','Marker border color. Example: "r", "[0.5,0,0]"'])
    dataInfoKeysGraphData.append(['markeredgewidth','Marker border width, in points. Example: "1.5"'])
    dataInfoKeysGraphData.append(['zorder',         'Determines the drawing order (float), highest is drawn on top.\nExample: "2", "3"'])
    dataInfoKeysGraphData.append(['== Offsets ==',  ''])
    dataInfoKeysGraphData.append(['offset',         'Offset to data. Examples: "-10" (on y values), or "[2,\'1/20\']" for (x,y) offset'])
    dataInfoKeysGraphData.append(['muloffset',      'Multiplitcative offset to data. Examples: "0.01" for y values, or\n"[10, 1e2]" for (x,y) multiplicative offsets'])
    dataInfoKeysGraphData.append(['== For specific curve types ==', ''])
    dataInfoKeysGraphData.append(['facecolor',      'Color of "fill" Curve types. Examples: "r", "[0.5,0,0]"'])
    dataInfoKeysGraphData.append(['cmap',           'Colormap, for Curve types which accept this keyword such as scatter). Examples:\n"afmhot", "inferno", or "[[0.91,0.25,1], [1.09,0.75,1], \'hls\']" (see Colorize options)'])
    dataInfoKeysGraphData.append(['vminmax',        'Bounds values for cmap. Examples: "[0,7]", or "[3, \'\']"'])
    dataInfoKeysGraphData.append(['colorbar',       'If not empty display a colorbar according to keyword cmap. Example: "1",\n"{\'ticks\': [-1, 0, 2]}", or "{\'orientation\':\'horizontal\', \'adjust\':[0.1, 0.1, 0.7, 0.1]}"'])
    dataInfoKeysGraphData.append(['xerr',           'x error for curve type "errorbar". Example: "1", or "5"'])
    dataInfoKeysGraphData.append(['yerr',           'y error for curve type "errorbar". Example: "1", or "5"'])
    dataInfoKeysGraphData.append(['== Misc ==',     ''])
    dataInfoKeysGraphData.append(['labelhide',      'Use "1" to hide label in the graph'])
    dataInfoKeysGraphData.append(['legend',         'Deprecated. Curve label in legend.'])
    dataInfoKeysGraphData.append(['ax_twinx',       'Plot curve on secondary axis. Example: "True", or anything'])
    dataInfoKeysGraphData.append(['ax_twiny',       'Plot curve on secondary axis. Example: "True", or anything'])
    dataInfoKeysGraphData.append(['linestyle',      'Use "none" to hide a Curve'])
    dataInfoKeysGraphData.append(["['key', value]", 'User-defined keyword-values pairs. Will be fed to the plotting method if possible.\nExamples: "[\'fillstyle\', \'top\']" for half-filled markers, "[\'comment\', \'a valuable info\']]'])
    dataInfoKeysGraphData_ = np.array([d[0:2] for d in dataInfoKeysGraphData])
    dataInfoKeysGraph = list(dataInfoKeysGraphData_[:,0])
    dataInfoKeysGraphExample = list(dataInfoKeysGraphData_[:,1])
    dataInfoKeysGraphExalist = []
    for i in range(len(dataInfoKeysGraphData)):
        if len(dataInfoKeysGraphData[i]) < 3:
            split = dataInfoKeysGraphData[i][1].split('"')
            dataInfoKeysGraphExalist.append(['']+split[1::2])
        else:
            dataInfoKeysGraphExalist.append(dataInfoKeysGraphData[i][2])
        
#    dataInfoKeysGraph = ['type', 'linespec', 'color', 'legend', 'label', 'labelhide', 'linewidth', 'facecolor', 'linestyle', 'alpha', 'marker', 'markersize', 'markerfacecolor', 'markeredgewidth', 'markeredgecolor', 'cmap', 'vminmax', 'colorbar', 'xerr', 'yerr', 'offset', 'muloffset', 'ax_twinx', 'ax_twiny', 'insetfile', "['key', value]"]
#    dataInfoKeysGraphExample = ['plot, scatter, fill, boxplot, errorbar, etc. (scatter_c after previous scatter)', '--r', 'r, [0.5,0,0], etc.', 'Some legend', 'Some legend', '1 to hide label in graph', '1.5', 'r, [0.5,0,0], etc.', '"" or none', '0.5, 1, etc.', 'o, or s, etc.', '2', 'r, [0.5,0,0], etc.', '1.', 'r, [0.5,0,0], etc.', 'afmhot, or inferno, or color list as in Colorize', 'Imposes bounds for cmap. Ex: [0,7], or [3, \'\']', '1, or {\'ticks\': [-1, 0, 2]}, or {\'orientation\':\'horizontal\', \'adjust\':[0.1, 0.1, 0.7, 0.1]}', '1, or 5. Requires \'type\': \'errorbar\'.', '1, or 5. Requires \'type\': \'errorbar\'.', '-10, or [2,20]', '0.01, or [10, 1e2]', 'True, or anything', 'True, or anything', 'file to place as inset. Related keywords: insetcoords and insetupdate', "['color', [0,0.2,0.7], or ['comment', 'a valuable info']]"]

    # constructor
    def __init__(self, filename='', complement='', silent=True, config='config.txt'):
        """ default constructor. Calls method reset. """
        # load preferences
        if config is not None:
            if config == 'config.txt':
                config = os.path.join(os.path.dirname(os.path.realpath(__file__)), config)
            self._config = Graph(config, complement={'readas': 'database'}, config=None)
        # actually load the file
        self.reset(filename, complement=complement, silent=silent)
    

    def reset(self, filename, complement='', silent=True):
        # complement: special keywords: 'readas', 'isfilecontent'
        # default values
        from grapa.graphIO import GraphIO
        self.filename = filename
        self.silent = silent
        self.data = []
        self.headers = {}
        self.graphInfo = {}
        self.sampleInfo = {}
        # try identify and parse the datafile
        if isinstance(filename, str) and filename == '':
            if not self.silent:
                print('Empty Graph object created')
            return
        if isinstance(filename, str):
            # if single file was provided - a string, or if filename is the content if the file - complement['isfilecontent'] must be true
            GraphIO.readDataFile(self, complement=complement)
        else:
            if len(filename) == 2 and len(filename[0]) == len(filename[1]) and is_number(filename[0][0]):
                # filename is actually the data, filename is a list
                if not self.silent:
                    print('Class Graph: interpret "filename" as data content')
                self.filename = ''
                GraphIO.dataFromVariable(self, filename, attributes=complement)
            else:
                # if a list was provided - open first file, then merge the others one by one
                if not isinstance(complement, list):
                    if complement != '':
                        print('WARNING class Graph: complement must be a list if filename is a list.')
                        print('   filename:', len(filename), 'element, complement:', complement)
                    complement = [complement] * len(filename)
                self.filename = filename[0]
                GraphIO.readDataFile(self, complement[0])
                if len(filename) > 1:
                    for i in range(1, len(filename)):
                        if len(complement) > 1:
                            self.merge(Graph(filename[i], complement=complement[i]))
                        else:
                            self.merge(Graph(filename[i]))
        # last: want to have abspath and not relative
        if hasattr(self, 'filename'):
            self.filename = os.path.abspath(self.filename)

# RELATED TO GUI
    def alterListGUI(self):
        out = []
        out.append(['Linear', ['', ''], '']) # do not want to leave this completely empty (if no curve left)
        for c in range(self.length()):
            for j in self.curve(c).alterListGUI():
                if j not in out:
                    out.append(j)
        return out




# "USUAL" CLASS METHODS
    def __str__(self):
        """ Returns some information about the class instance. """
        out = 'Content of Graph stored in file ' + self.filename + '\n'
        out += 'Content of headers:' + '\n'
        out += str(self.headers)
        out += '\nNumber of Curves: '+str(self.length())
        return out


    # interactions with other Graph objects
    def merge(self, graph):
        """ Add in a Graph object the content of another Graph object """
        # keep previous self.filename
        # copy data
        for x in graph.data:
            self.data.append(x)
        # copy headers, unless already exists (in this case information is forgotten)
        for key in graph.headers:
            if not key in self.headers:
                self.headers.update({key: graph.headers[key]})
        # copy graphInfo, unless already exists (in this case information is forgotten)
        for key in graph.graphInfo:
            if not key in self.graphInfo:
                self.graphInfo.update({key: graph.graphInfo[key]})
        # copy sampleInfo, unless already exists (in this case information is forgotten)
        for key in graph.sampleInfo:
            if not key in self.sampleInfo:
                self.sampleInfo.update({key: graph.sampleInfo[key]})

    
    # methods handling the bunch of curves
    def length(self):
        """ Returns the number of Curve objects in the list. """
        return len(self.data)

    def curve(self, index):
        """ Returns the Curve object at index i in the list. """
        if index >= self.length() or self.length() == 0:
            print ('ERROR Class Graph method Curve: cannot find Curve (index',index,', max possible',self.length()-1,')')
            return 
        return self.data[index]

    def curves(self, attr, value, strLower=False, strStartWith=False):
        """ 
        Returns a list of Curves which attribute attr == value (also same type)
        attr: the attribute to check. By default 'label'.
        strLower: if True, 
        strStartWith: is both value and the attribute are str, only look at the
            first characters
        """
        out = []
        for c in self.iterCurves():
            lbl = c.getAttribute(attr)
            if isinstance(lbl, type(value)):
                if strLower:
                    if isinstance(lbl, str):
                        lbl = lbl.lower()
                    if isinstance(value, str):
                        value = value.lower()
                if lbl == value:
                    out.append(c)
                elif isinstance(lbl, str) and strStartWith and lbl[:len(value)] == value:
                    out.append(c)
        return out

    def iterCurves(self):
        """ Returns an iterator over the different Curves """
        for c in range(self.length()):
            yield self.curve(c)
    
    def append(self, curve):
        """
        Add into the object list a Curve, a list or Curves, or every Curve in a
        Graph object.
        """
        if isinstance(curve, Graph):
            for c in curve.iterCurves():
                self.data.append(c) # has to be Curves
        if isinstance(curve, list):
            for c in curve:
                self.append(c) # call itself, must check if c is a Curve
        elif isinstance(curve, Curve):
            self.data.append(curve)
        elif isinstance(curve, str) and curve == 'empty':
            self.data.append(Curve([[np.inf],[np.inf]], {}))
        else:
            print('Graph.append: failed (type:',type(curve),')')

    def deleteCurve(self, i):
        """ Delete a Curve at index i from the Graph object. """
        if isinstance(i, (list, tuple)):
            i = list(np.sort(i)[::-1])
            for k in i:
                self.deleteCurve(k)
        else:
            le0 = len(self.data)
            if is_number(i) and i < le0:
                # delete data
                del self.data[i]
                # delete in headers
                if 'collabels' in self.headers and i < len(self.headers['collabels']):
                    del self.headers['collabels'][i]
                if 'collabelsdetail' in self.headers: # certainly only useful for "databases"
                    for j in range(len(self.headers['collabelsdetail'])):
                        if i < len(self.headers['collabelsdetail'][j]):
                            del self.headers['collabelsdetail'][j][i]
                # nothing to delete in graphInfo
                # nothing to delete in sampleInfo

    def replaceCurve(self, newCurve, idx):
        """ Replaces a Curve with another. """
        if isinstance(newCurve, Curve):
            try:
                self.data[idx] = newCurve
                return True
            except Exception:
                print ('Graph.replaceCurve: cannot add Curve at index', idx, '.')
        else:
            print ('Graph.replaceCurve: newCurve is not a Curve (type', type(newCurve), ')')
        return False
    
    def swapCurves(self, idx1, idx2, relative=False):
        """
        Exchange the Curves at index idx1 and idx2.
        For example useful to modify the plot order (and order in the legend)
        """
        if idx1 < -self.length() or idx1 >= self.length():
            print ('Graph.swapCurves: idx1 not valid (value', idx1, ').')
            return False
        if relative:
            idx2 = idx1 + idx2
        if idx2 == idx1: # swap with itself
            return True
        if idx2 < -self.length() or idx2 >= self.length():
            print ('Graph.swapCurves: idx2 not valid (value', idx2, ').')
            return False
        swap = deepcopy(self.curve(idx1))
        self.data[idx1] = self.curve(idx2)
        self.data[idx2] = swap
        return True

    def duplicateCurve(self, idx1):
        """ Duplicate (clone) an existing curve and append it in the curves list."""
        if idx1 < -self.length() or idx1 >= self.length():
            print ('Graph.duplicateCurve: idx1 not valid (value', idx1, ').')
            return False
        curve = deepcopy(self.curve(idx1))
        self.data.insert(idx1+1, curve)

        
    # methods handling content of curves
    def getCurveData(self, idx):
        alter = self._getAlter()
        x = self.curve(idx).x(alter=alter[0])
        y = self.curve(idx).y(alter=alter[1])
        return np.array([x,y])

    def update(self, attributes, ifAll=False, forceGraphInfo=False):
        """ Update the properties of the Graph object.
        The properties are stored in dict objects, so update works similarly as
        the dict.update() method.
        If a property is not known to belong to the Graph object, it is
        attributed to the 1st Curve object (unless ifAll=True), or if
        forceGraphInfo is True (then the undetermined keys are inserted in
        self.graphInfo)
        """
        for key in attributes:
            k = key.lower().replace('ï»¿','')
            try:
                if   k in self.headersKeys:
                    if attributes[key] != '':
                        self.headers.update({k: attributes[key]})
                    elif k in self.headers:
                        del self.headers[k]
                elif (k in self.graphInfoKeys or forceGraphInfo) or k.startswith('subplots'):
                    if attributes[key] is not '':
                        self.graphInfo.update({k: attributes[key]})
                    elif k in self.graphInfo:
                        del self.graphInfo[k]
                    # by default nothing in sampleInfo, everything in the curves
                else:
                    if ifAll:
                        for i in range(self.length()):
                            self.curve(i).update({k: attributes[key]})
                    else:
                        self.curve(-1).update({k: attributes[key]})
            except Exception as e:
                print('Error Graph.update: key', key, ' attributes', attributes, 'exception', e)

    def updateValuesDictkeys(self, *args, **kwargs):
        """
        Performs update({key1: value1, key2: value2, ...}) with
        arguments value1, value2, ... , (*args) and
        kwargument key=['key1', 'key2', ...]
        """
        if 'keys' not in kwargs:
            print('Error Graph updateValuesDictkeys: "keys" key must be provided, must be a list of keys corresponding to the values provided in *args.')
            return False
        if len(kwargs['keys']) != len(args):
            print('WARNING Graph updateValuesDictkeys: len of list "keys" argument must match the number of provided values (',len(args),' args provided, ',len(kwargs['keys']),' keys).')
        lenmax = min(len(kwargs['keys']), len(args))
        for i in range(lenmax):
            self.update({kwargs['keys'][i]: args[i]})
        return True
                
    def delete(self, key, ifAll=False):
        """
        Delete an attribute of the Graph if recognized as such, or of the curve.
        Return the deleted attribute (except for attributes of curves if ifAll=True)
        """
        k = key.lower()
        out = {}
        if k in self.headersKeys:
            if k in self.headers:
                out.update({key: self.headers[k]})
                del self.headers[k]
        elif k in self.graphInfoKeys:
            if k in self.graphInfo:
                out.update({key: self.graphInfo[k]})
                del self.graphInfo[k]
        else:
            if ifAll:
                for i in range(self.length()):
                    # cannot export in this case
                    self.data[i].delete(key)
            else:
                out.update(self.curve(-1).delete(key))
        return out

    def getAttribute(self, attr, default=''):
        at = attr.lower()
        if at in self.headers:
            return self.headers[at]
        if at in self.graphInfo:
            return self.graphInfo[at]
        if at in self.sampleInfo:
            return self.sampleInfo[at]
        if self.length() > 0:
            return self.curve(0).getAttribute(attr, default=default)
        return default

    def deleteAttr(self, attrList):
        out = {}
        for key in attrList:
            out.update({key: self.getAttribute(key)})
            self.delete(key)
        return out

    def castCurve(self, newtype, idx, silentSuccess=False):
        """
        Replace a Curve with another type of Curve with identical data and 
        properties.
        """
        if idx >= - self.length() and idx < self.length():
            newCurve = self.curve(idx).castCurve(newtype)
            if isinstance(newCurve, Curve):
                flag = self.replaceCurve(newCurve, idx)
                if flag:
                    if not silentSuccess:
                        print ('Graph.castCurve: new Curve type:', self.curve(idx).classNameGUI() + '.')
                else:
                    print ('Graph.castCurve')
                return flag
        else:
            print ('Graph.castCurve: idx not in suitable range (', idx,', max', self.length(),').')
        return False
    

    def colorize(self, colorscale, sameIfEmptyLabel=False, avoidWhite=False):
        """
        Colorize a graph, by coloring each Curve along a colorscale gradient.
        """
        from grapa.colorscale import Colorscale
        if not isinstance(colorscale, Colorscale):
            colorscale = Colorscale(colorscale)
        if self.length() == 1:
            self.curve(0).update({'color': colorscale.valuesToColor(0.0, avoidWhite=avoidWhite)})
            return
        show = np.arange(0, self.length())
        if sameIfEmptyLabel:
            show = np.array([0.0] * self.length())
            i = 0.0
            for c in range(self.length()):
                show[c] = i
                if self.curve(c).getAttribute('label') != '' and not self.curve(c).isHidden():
                    i += 1.0
                elif c > 0:
                    show[c] = show[c-1]
        cols = colorscale.valuesToColor(show/max(max(show),1.0), avoidWhite=avoidWhite)
        for c in range(self.length()):
            self.curve(c).update({'color': cols[c]})


    def applyTemplate(self, graph, alsoCurves=True):
        """
        Apply a template to the Graph object. The template is a Graph object
        which contains:
        - self.getAttributes listed in graphInfoKeys
        - self.curve(i).getAttributes listed in dataInfoKeysGraph
        alsoCurves: True also apply Curves properties, False only apply Graph
        properties
        """
        # strip default attributes
        for key in Graph.DEFAULT:
            if graph.getAttribute(key) == Graph.DEFAULT[key]:
                graph.update({key: ''})
        for key in graph.graphInfoKeys:
            val = graph.getAttribute(key)
            if val != '':
                self.update({key: val})
        if alsoCurves:
            for c in range(self.length()):
                if c >= graph.length():
                    break
                for key in self.dataInfoKeysGraph:
                    val = graph.curve(c).getAttribute(key)
                    if val != '':
                        self.curve(c).update({key: val})
    
    def replaceLabels(self, old, new):
        """
        Modify all labels of the Graph, by replacing 'old' by 'new'
        """
        for c in self.iterCurves():
            c.update({'label': c.getAttribute('label').replace(old, new)})

    def checkValidText(self):
        text = self.getAttribute('text', None)
        texy = self.getAttribute('textxy', '')
        targ = self.getAttribute('textargs', {})
        if text == None:
            self.update({'textxy': '', 'textargs': ''})
            return
        onlyfirst = False if isinstance(text, list) else True
        # transform everything into lists
        if not isinstance(text, list):
            text = [text]
        if not isinstance(targ, list):
            targ = [targ]
        if not isinstance(texy, list):
            texy = [texy]
        if len(texy) == 2 and isinstance(texy[0], (list, tuple)):
            texy = [texy]  # if texy was like (0.5,0.8)
        for i in range(len(targ)):
            if not isinstance(targ[i], dict):
                #print('Graph.checkValidText targ set', i, '{} (previous', targ[i], ')')
                targ[i] = {}
        for i in range(len(texy)):
            if not isinstance(texy[i], (tuple, list)):
                #print('Graph.checkValidText texy set', i, '\'\' (previous', texy[i], ')')
                texy[i] = ''
        while len(texy) < len(text):
            texy.append(texy[-1])
        while len(targ) < len(text):
            targ.append(targ[-1])
        if onlyfirst:
            text, texy, targ = text[0], texy[0], targ[0]
        if text != self.getAttribute('text'):
            print('Corrected attribute text', text, '(former', self.getAttribute('text'),')')
        if texy != self.getAttribute('textxy'):
            print('Corrected attribute textxy', texy, '(former', self.getAttribute('textxy'),')')
        if targ != self.getAttribute('textargs'):
            print('Corrected attribute textargs', targ, '(former', self.getAttribute('textargs'),')')
        self.update({'text': text, 'textxy': texy, 'textargs': targ})
                    
    def addText(self, text, textxy, textargs=None):
        """
        Adds a text to be annotated in the plot, handling the not-so-nice
        internal implementation
        """
        if textargs is None:
            textargs = {}
        restore = []
        restore.append({'text':     self.getAttribute('text'),
                        'textxy':   self.getAttribute('textxy'),
                        'textargs': self.getAttribute('textargs')})
        attrs = {'text': text, 'textxy': textxy, 'textargs': textargs}
        # TODO: ensure text, textxy and textargs if list have same length
        if self.getAttribute('text', None) is None:
            self.update(attrs)
        else:
            if not isinstance(self.getAttribute('text'), list):
                for key in attrs.keys(): # ensures attributes are list
                    self.update({key: [self.getAttribute(key, '')]})
            for key in attrs.keys():
                self.update({key: self.getAttribute(key) + [attrs[key]]})
        return restore
    def removeText(self):
        restore = []
        restore.append({'text':     self.getAttribute('text'),
                        'textxy':   self.getAttribute('textxy'),
                        'textargs': self.getAttribute('textargs')})
        attrs = ['text', 'textxy', 'textargs']
        if self.getAttribute('text', None) is None:
            pass
        else:
            if not isinstance(self.getAttribute('text'), list):
                for key in attrs: # ensures attributes are list
                    self.update({key: [self.getAttribute(key, '')]})
            for key in attrs:
                self.update({key: self.getAttribute(key)[:-1]})
        return restore

            
            
    
    # some mathemetical operation on curves
    def curvesAdd(self, idx0, idx1, interpolate=0, **kwargs):
        """ addition operation when only knowing curves indices """
        kwargs.update({'interpolate': interpolate, 'sub': False})
        return self.curve(idx0).__add__(self.curve(idx1), **kwargs)
    def curvesSub(self, idx0, idx1, interpolate=0, **kwargs):
        """ substraction operation when only knowing curves indices """
        kwargs.update({'interpolate': interpolate, 'sub': True})
        return self.curve(idx0).__add__(self.curve(idx1), **kwargs)
        
    def _curveMethod_graphRef(self, *args, **kwargs):
        """
        Allows operations on the graph to be executed from inside the Curve
        object. See for example CurveMath.
        Expect argument 'curve': a Curve object
        Expect argument 'method': will call curve.method
        """
        if 'curve' in kwargs and 'method' in kwargs:
            curve = kwargs['curve']
            m = kwargs['method']
            del kwargs['curve'], kwargs['method']
            return getattr(curve, m)(*args, graph=self, **kwargs)

    
    def _getAlter(self):
        """ returns the formatted alter instruction of self """
        return self._getAlterToFormat(self.getAttribute('alter'))
    @classmethod
    def _getAlterToFormat(cls, alter):
        """ Return a formatted alter instruction """
        if alter == '':
            alter = ['', '']
        if isinstance(alter, str): # nothing to do if it is dict
            alter = ['', alter]
        return alter
        

            
    def formatAxisLabel(self, label):
        """
        Returns a string for label according to user preference.
        'Some quantity [unit]' may become 'Some quantify (unit)'
        Other possible input is ['This is a Quantity', 'Q', 'unit']
        """
        # retrieve user preference
        symbol= bool(self.config('graph_labels_symbols', False))
        units = self.config('graph_labels_units', default='[]', astype=str)
        units = units.replace('unit', '').replace(' ','')
        # format input
        if isinstance(label, str):
            if units != '[]': # that is default, no need to do anything
                expr = '^(.* )\[(.*)\](.*)$'
                if label != '':
                    f = refindall(expr, label)
                    if isinstance(f, list) and len(f) == 1 and len(f[0]) == 3:
                        if units in ['DIN', '/']:
                            return f[0][0].strip(' ') + ' / ' + f[0][1] + '' + f[0][2]
                        elif units == '()':
                            return f[0][0].strip(' ') + '(' + f[0][1] + ')' + f[0][2]
        elif isinstance(label, list):
            while len(label) < 3:
                label += ['']
            out = label[0]
            if symbol and len(label[1]) > 0:
                out += ' $' + label[1] + '$'
            if label[2] not in [None, ' ']:
                if units == '/':
                    out += ' / ' + label[2]
                elif units == '()':
                    out += ' (' + label[2] + ')'
                else:
                    out += ' [' + label[2] + ']'
            return out.replace('  ', ' ')
        return label
        
        
    def config(self, key, default='', astype='auto'):
        """
        returns the value corresponding to key in the configuration file.
        If key is not defined, returns default.
        """
        if hasattr(self, '_config') and self._config.length() > 0:
            out = self._config.curve(0).getAttribute(key, default=default)
            if astype in [str, 'str']:
                return str(out)
            else:
                return stringToVariable(out)
        return default

    
    def filenamewithpath(self, filename):
        # if relative, join the path of the file with
        if os.path.isabs(filename):
            return filename
        path = ''
        if hasattr(self, 'filename') and isinstance(self.filename, str) and len(self.filename) > 0:
            path = os.path.dirname(os.path.abspath(self.filename))
        return os.path.join(path, filename)
        
        
    # For convenience we offer a shortcut to GraphIO.export
    def export(self, filesave='', saveAltered=False, ifTemplate=False,
               ifCompact=True, ifOnlyLabels=False, ifClipboardExport=False):
        """
        Exports content of Grah object into a human- and machine- readable
        format.
        """
        from grapa.graphIO import GraphIO
        return GraphIO.export(self, filesave=filesave, saveAltered=saveAltered,
                              ifTemplate=ifTemplate, ifCompact=ifCompact,
                              ifOnlyLabels=ifOnlyLabels,
                              ifClipboardExport=ifClipboardExport)
        

    # For convenience we offer a shortcut to GraphIO.plot
    def plot(self, filesave='', imgFormat='', figsize=(0, 0), ifSave='auto',
             ifExport='auto', figAx=None, ifSubPlot='auto'):
        """
        Plot the content of the object.
        filesave: filename for the saved graph image and export text
            file.
        imgFormat: by default image format will be .png. Possible formats are
            the ones supported by plt.savefig() and possibly .emf
        figsize: figure size in inch. The actual default value is from a class
            constant
        ifSave: [True/False/'auto'] if True, save the Graph as an image. If
            left to default, saves the image only if filesave is provided
        ifExport: [True/False/'auto'] if True, create a human- and machine-
            readable .txt file containing all information of the graph.
            By default, only export if filesave is provided
        figAx as [fig, ax]: the graph will be plotted in the provided figure
            and ax.
            Providing [fig, None] will erase all existing axes and create what
            is needed
        ifSubPlot: [True/False/'auto'] If True, prevents deletion of the
            existing axes in the figure.
            By default keeps axes if axes are provided in figAx.
            
        """
        if ifSave not in [True, False]:
            ifSave = False if filesave == '' else True
        if ifExport not in [True, False]:
            ifExport = False if filesave == '' else True
        from grapa.graphIO import GraphIO
        if ifSubPlot not in [True, False]:
            ifSubPlot = False
            if (figAx is not None and isinstance(figAx, list)
                and len(figAx) > 1 and figAx[1] is not None):
                ifSubPlot = True
        return GraphIO.plot(self, filesave=filesave, imgFormat=imgFormat, figsize=figsize,
                            ifSave=ifSave, ifExport=ifExport, figAx=figAx,
                            ifSubPlot=ifSubPlot)



# -*- coding: utf-8 -*-
"""
Created on Sat Mar  4 10:57:14 2017

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""


from grapa.graph import Graph
from grapa.graphIO import GraphIO
from grapa.mathModule import is_number


class GraphXLS(Graph):
    
    FILEIO_GRAPHTYPE = 'Undetermined measurement type'
    
    @classmethod
    def isFileReadable(cls, fileName, fileExt, **kwargs):
        if fileExt in ['.xls', '.xlsx']:
            return True
        return False

    def readDataFromFile(self, attributes, **kwargs):
        try:
            from xlrd import open_workbook
        except ImportError as e:
            print('ImportError: cannot import xlrd. GraphXLS aborted.', e)
            return False
    
        wb = open_workbook(self.filename)
        sheet_id = attributes['complement'] if 'complement' in attributes else 0
        if not isinstance(sheet_id, str) and not is_number(sheet_id):
            sheet_id = attributes['sheetid'] if 'sheetid' in attributes else 0
        try: # first try by name, then try by index, try first sheet by default
            sheet = wb.sheet_by_name(sheet_id)
        except Exception:
            try:
                sheet = wb.sheet_by_index(sheet_id)
            except Exception:
                sheet = wb.sheet_by_index(0)
                print('readDataFromFileXLS: spreadsheet "', sheet_id, '" not found. Opened first sheet by default.')
        number_of_rows = sheet.nrows
        number_of_columns = sheet.ncols
        items = []
        for row in range(0, number_of_rows):
            values = []
            for col in range(number_of_columns):
                try: # required to process correctly rows with merged cells
                    value = (sheet.cell(row, col).value)
                except Exception:
                    value = ''
                try:
                    value = str(float(value))
                except ValueError:
                    pass
                finally:
                    values.append(value)
            items.append(values)
        # identify correct way to process data - is it database, or file generic ?
        func = GraphIO._funcReadDataFile(self, items, attributes)
        func(self, attributes, fileContent=items)
        


# -*- coding: utf-8 -*-
"""
Created on Fri Jan 26 13:30:22 2018

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""


import numpy as np
from grapa.graph import Graph
from grapa.curve import Curve


class GraphPGZ(Graph):
    
    FILEIO_GRAPHTYPE = 'PGZ path file'
    
    AXISLABELS = [['Position x', '', 'mm'], ['Position y', '', 'mm']]
    
    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', **kwargs):
        if (fileExt == '.pgz'):
            return True
        return False

    def readDataFromFile(self, attributes, **kwargs):
        kw = {'delimiter': '\t', 'invalid_raise': False,
              'skip_header': 0, 'usecols': [0, 1, 2]}
        data = np.genfromtxt(self.filename, **kw)
        
        # transform data format
        x = [[], []]
        y = [[], []]
        previous = -1
        for row in data:
            onOff = 1 if row[2] == 1 else 0
            if row[2] != previous and len(x[1-onOff]) > 0:
                x[onOff].append(x[1-onOff][-1])
                y[onOff].append(y[1-onOff][-1])
                x[1-onOff].append(np.nan)
                y[1-onOff].append(np.nan)
                #print('   added from other')
            x[onOff].append(row[0])
            y[onOff].append(row[1])
            #print('   added to', onOff)
            previous = row[2] * 1
            
        self.append(Curve([x[0], y[0]], attributes))
        self.append(Curve([x[1], y[1]], attributes))
        lbl = self.curve(-2).getAttribute('label')
        self.curve(-1).update({'label': lbl+' off', 'labelhide': 1, 'alpha':0.20})
        
        colors = {'P1': [0.3,0.3,0.3], 'P2': [1,0,0], 'P3': [0,0,1], 'PT':[0,0.6,0.6]}
        color = ''
        for key in colors:
            if key in lbl:
                color = colors[key]
                break
        if color != '':
            self.curve(-2).update({'color': color})
            self.curve(-1).update({'color': color})
            
        # some default settings
        self.update({'figsize':[5,5], 'subplots_adjust': [0.15, 0.15, 0.95, 0.95]})
        self.update({'xlabel': self.formatAxisLabel(GraphPGZ.AXISLABELS[0]),
                     'ylabel': self.formatAxisLabel(GraphPGZ.AXISLABELS[1])})




# -*- coding: utf-8 -*-
"""
Created on Fri Dec 23 16:46:38 2016

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import glob
import os
import copy
import numpy as np
import matplotlib.pyplot as plt
import sys

path = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', '..'))
if path not in sys.path:
    sys.path.append(path)

from grapa.graph import Graph
from grapa.curve import Curve
try:
    from grapa.scripts.script_processJV import writeFileAvgMax
except ImportError:
    pass



def replaceLabels(graph, label):
    print('label', label)
    if label in ['Voc [V]']:
        label = graph.formatAxisLabel(['Voc', '', 'V'])
    elif label in ['Jsc [mA/cm2]']:
        label = graph.formatAxisLabel(['Jsc', '', 'mA cm$^{-2}$'])
    elif label in ['FF [%]']:
        label = graph.formatAxisLabel(['FF', '', '%'])
    elif label in ['Eff. [%]']:
        label = graph.formatAxisLabel(['Eff.', '', '%'])
    elif label in ['Rp [Ohmcm2]']:
        label = graph.formatAxisLabel(['Rp', '', 'Ohm cm$^2$'])
    elif label in ['Rs [Ohmcm2]']:
        label = graph.formatAxisLabel(['Rs', '', 'Ohm cm$^2$'])
    elif label in ['n']:
        label = graph.formatAxisLabel(['n', '', ' '])
    elif label in ['J0 [A/cm2]']:
        label = graph.formatAxisLabel(['J$_0$', '', 'A cm$^{-2}$'])
    return label

def labelFromGraph(graph, silent=False, file=''):
    # label will be per priority:
    # 1. if a 'label' field is present at beginning of file (so all curve labels are identical)
    # 2. a 'sample' field of 'sample name' field at beginning of file 
    # 3. processed filename
    
    # if label set at beginning of the file -> same label for each curve
    lbl = graph.curve(0).getAttribute('label')
    for c in range(graph.length()):
        if graph.curve(c).getAttribute('label') != lbl:
            lbl = None
            break
    if lbl is not None:
        if not silent:
            print('label:', lbl, '(\'label\' attribute in file header, or unexpected input)')
    if lbl is None:
        # if not every labels are identical: look for 'sample' and 'sample name' fields
        # overrides if find 'sample' keyword in the headers of the file
        if graph.curve(0).getAttribute('sample name') != '':
            lbl = graph.curve(0).getAttribute('sample name')
            if isinstance(lbl, list):
                lbl = lbl[0]
                for c in range(graph.length()): # clean a bit the mess
                    if isinstance(graph.curve(c).getAttribute('sample name'), list):
                        graph.curve(c).update({'sample name': graph.curve(c).getAttribute('sample name')[0]})
            if not silent:
                print('label:', lbl, '(\'sample name\' attribute in file header)')
        elif graph.getAttribute('sample') != '':
            lbl = graph.getAttribute('sample')
            if not silent:
                print('label:', lbl, '(\'sample\' attribute in file header)')
    if lbl is None:
        lbl = os.path.basename(file).replace('export','').replace('summary','').replace('_',' ').replace('  ',' ')
        if not silent:
            print('label:', lbl, '(from file name - no \'label\', \'sample\' or \'sample name\' attribute found in file header)')
    return lbl

    
    
    
def JVSummaryToBoxPlots(folder='.', exportPrefix='boxplot_', replace=None, plotkwargs={}, silent=False, pltClose=True, newGraphKwargs={}):
    if replace is None:
        replace = [] # possible input: [['Oct1143\n', ''], ['\nafterLS','']]
    newGraphKwargs =  copy.deepcopy(newGraphKwargs)
    newGraphKwargs.update({'silent': silent})
    
    # establish list of files
    path = os.path.join(folder, '*.txt')
    fileList = sorted(glob.glob(path))
    files = []
    for file in fileList:
        filename = os.path.basename(file)
        # do notwant to open previously created export files
        if len(exportPrefix) > 0 and filename[:len(exportPrefix)] != exportPrefix:
            files += [file]
    if len(files) == 0:
        return None
    
    # default mode: every column is shown, grouped by columns
    mode = 'all'
    # test if JV mode: only a subset is shown, and some cosmetics is performed
    isJV = True
    graph = Graph(files[0], **newGraphKwargs)
    collabels = graph.getAttribute('collabels')
    if len(collabels) < 8:
        isJV = False
    else:
        collabels = graph.getAttribute('collabels')
        cols = [0, 1, 2, 3, 8, 9] # required columns
        titles = [['Voc'], ['Jsc'], ['FF'], ['Eff'], ['Rp'], ['Rs']]
        for c in range(len(cols)):
            flag = False
            for tmp in titles[c]:
                if tmp in collabels[cols[c]]:
                    flag = True
                    break
            if not flag:
                isJV = False
    if isJV:
        mode = 'JV'
        JVcols = [0, 1, 2, 3, 8, 9, 10, 11] # all columns of interest
        JVtitles = [['Voc'], ['Jsc'], ['FF'], ['Eff'], ['Rp'], ['Rs'], ['n'], ['J0','I0']]
        JVupdates = [{}] * len(JVcols)
        JVupdates[-1] = {'typeplot': 'semilogy'} # J0

    graphs = []
    titles = []
    
    nbCarriageReturnLabel = 0
    strStatistics = ''
    
    for file in files :
        # open the file corresponding to 1 given sample
        if not silent:
            print ('open file',file)
        graph = Graph(file, **newGraphKwargs)
        if graph.length() == 0:
            print('Cannot interpret file data. Go to next one.')
            continue
        
        complement = copy.deepcopy(graph.curve(0).getAttributes())
        complement.update({'type': 'boxplot'})
        
        if mode == 'JV':
            # gather statisics data to be placed in a text file
            try:
                strStatistics += writeFileAvgMax(graph, withHeader=(True if len(strStatistics) == 0 else False), colSample=True, filesave=None)
            except Exception as e:
                print('Exception += ', type(e), e)
                pass
        
        label = labelFromGraph(graph, silent=silent, file=file)
        # "clean" label according to replacement pairs provided by user
        for rep in replace:
            label = label.replace(rep[0], rep[1])
        complement.update({'label': label})
        
        # needed to estimate the graph height
        nbCarriageReturnLabel = max(nbCarriageReturnLabel, label.count('\n'))
    
        # construct one column of each box plot
        collabels = graph.getAttribute('collabels')
        curvesOfInterest = JVcols if mode == 'JV' else range(graph.length())
        for i in range(len(curvesOfInterest)):
            c = curvesOfInterest[i]
            data = graph.curve(c).getData()
            
            # prepare graph title
            while i >= len(titles):
                titles.append('')
            title = collabels[c] if titles[i] == '' else ''
                
            # if input not match JV file, discard it
            if mode == 'JV':
                flag = False
                if c < len(collabels):
                    for tit in JVtitles[i]:
                        if collabels[c].startswith(tit):
                            flag = True
                            break
                if not flag:
                    print('Wrong column name, no data plotted, file', file, 'i', i, 'column', c)
                    data = [[np.nan], [np.nan]]
                    title = ''
                title = title.replace('Voc_V','Voc [V]').replace('Jsc_mApcm2','Jsc [mA/cm2]').replace('FF_pc','FF [%]').replace('Eff_pc','Eff. [%]').replace('_Ohmcm2',' [Ohmcm2]')
                title = title.replace('A/cm2', 'A cm$^{-2}$').replace('J0 ','J$_0$ ').replace('Ohmcm2', '$\Omega$ cm$^2$')
            
            # prepare graph
            if title is not '':
                titles[i] = graph.formatAxisLabel(title)
            while i >= len(graphs):
                graphs.append(Graph('', **newGraphKwargs))
            graphs[i].append(Curve(data, complement))
        # go to next file
        
    # cosmetics, then save plots
    figSize = (1 + 0.9*len(files), 4)
    for i in range(len(graphs)):
        graph = graphs[i]
        if graph.length() > 0:
            # counts number of non-NaN element in each curve, only plot if is > 0
            num = sum([sum(~np.isnan(c.x())) for c in graph.iterCurves()])
            if num > 0:
                graph.update({'figsize': figSize, 'subplots_adjust': [1.1 / figSize[1], (nbCarriageReturnLabel+2)*0.06]})
                #graph.update({'title': titles[i]})
                graph.update({'ylabel': titles[i]})
                filesave = os.path.join(folder, exportPrefix + str(i))
                if mode == 'JV':
                    if JVupdates[i] is not {}:
                        graph.update(JVupdates[i])
                    tit = titles[i]
                    filesave = os.path.join(folder, exportPrefix + tit.split(' ')[0].replace('$','').replace('.','').replace('_',''))
                graph.plot(filesave=filesave, **plotkwargs)
                if pltClose:
                    plt.close()
    
    # print and save statistics
    if len(strStatistics) > 0:
        print(strStatistics)
        filesave = os.path.join(folder, exportPrefix + 'statistics.txt')
        print('filesave', filesave)
        f = open(filesave, 'w')
        f.write(strStatistics)
        f.close()

    # return last plot
    print('JVSummaryToBoxPlots completed.')
    return graphs[0]


if __name__ == "__main__":
    folder = './../examples/boxplot/'
#    folder = './../examples/boxplot/notJVspecific/'
    JVSummaryToBoxPlots(folder=folder, exportPrefix='boxplots_', pltClose=False, silent=True)
    plt.show()




# -*- coding: utf-8 -*-
"""
Created on Fri Jul 15 15:46:13 2016

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import numpy as np
import os
from scipy import interpolate


from grapa.graph import Graph
from grapa.curve import Curve
from grapa.mathModule import roundSignificant



class GraphEQE(Graph):
    
    FILEIO_GRAPHTYPE = 'EQE curve'
    FILEIO_GRAPHTYPE_OLD = 'EQE curve (old)'
    
    AXISLABELS = [['Wavelength', '\lambda', 'nm'], ['Cell EQE', '', '%']]
    
    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', **kwargs):
        # new QE files
        if fileExt == '.sr' and line1 == 'This is a QE measurment':
            return True
        # old setup QE files (2013(?) or older)
        if fileExt == '.sr' and line1[0:16] == 'Reference Cell: ':
            return True
        return False

    def readDataFromFile(self, attributes, **kwargs):
        ifOld = False
        # retrieve sample name - code should work for old and new file format
        f = open(self.filename, 'r')
        line = f.readline() # retrieve sample, should be stored on the 2nd line
        if line[0:16] == 'Reference Cell: ':
            ifOld = True
        line = list(filter(len,
                f.readline().replace(': ', '\t').strip(' \r\n\t').split('\t')))
        # look for acquisition attributes
        attributesFile = {}
        interests = [['Amplification settings', 'amplification'],
                     ['Lockin settings', 'lockin']]
        for l in f:
            for interest in interests:
                if l.startswith(interest[0]):
                    for param in l.strip().split('\t'):
                        if ': ' in param:
                            tmp = param.split(': ')
                            attributesFile.update({interest[1]+tmp[0].replace(' ',''): tmp[1]})
            if len(l) == 1 and l[0] in ['\n']:
                break
        f.close()
        data = np.array([np.nan, np.nan])
        try:
            kw = {'delimiter': '\t', 'invalid_raise': False}
            if ifOld:
                kw.update({'skip_header': 14, 'usecols': [0, 6]})
            else:
                kw.update({'skip_header': 15, 'usecols': [0, 5]})
            data = np.transpose(np.genfromtxt(self.filename, **kw))
        except Exception:
            if not self.silent:
                print('readDataFromFileEQE cannot read file', self.filename)
        # normalize data
        self.append(CurveEQE(data, attributes))
        self.curve(-1).update(attributesFile)
        if len(data.shape) > 1:
            self.curve(-1).update({'mulOffset': 100})
        # update label with information stored inside the file
        self.update({'label': line[-1].replace('_', ' ')})
        # some default settings
        self.update({'xlabel': self.formatAxisLabel(GraphEQE.AXISLABELS[0]),
                     'ylabel': self.formatAxisLabel(GraphEQE.AXISLABELS[1]),
                     'ylim': [0, 100], 'xlim': [300, np.nan]})
        self.headers.update({'collabels': ['Wavelength [nm]', 'EQE [%]']})
        # newer versions of EQE include phase signal -> one more header line
        self.update({'sample': self.curve(0).getAttribute('label')})
        # default value in files old setup
        if self.headers['sample'] == 'Ref':
            filenam_, fileext = os.path.splitext(self.getAttribute('filename'))
            self.curve(-1).update({'label':
                                   (filenam_.split('/')[-1]).split('\\')[-1]})
            self.headers.update({'sample': filenam_})
        if ifOld:
            self.headers.update({'meastype': GraphEQE.FILEIO_GRAPHTYPE_OLD})



    

class CurveEQE(Curve):
    
    CURVE = 'Curve EQE'

    EQE_AM15_REFERENCE = 'AM1-5_Ed2-2008.txt'
    EQE_AM0_REFERENCE = 'AM0_2000_ASTM_E-490-00.txt'
    EQE_BEST_CELL_REF = 'EQE_20.4_cell.txt'
    EQE_BEST_CELL_LABEL = 'Empa 20.4%'

    
    def __init__(self, data, attributes, silent=False):
        # modify default label
        if 'label' in attributes:
            if not isinstance(attributes['label'], str):
                attributes['label'] = str(attributes['label'])
            if '$' not in attributes['label']:
                attributes['label'] = attributes['label'].replace('_', ' ')
        if not isinstance(data, (np.ndarray)):
            data = np.array(data)
        # detect possible empty lines at beginnning (some versions of EQE files
        # ahave different number of header lines)
        flag = True if len(data.shape) > 1 else False # check if data loaded
        while flag:
            if data[0, 0] == 0:
                data = data[:, 1:]
            else:
                flag = False
        # main constructor
        Curve.__init__(self, data, attributes, silent=silent)
        # bandgap calculation
        popt = self.getAttribute('_popt')
        if isinstance(popt, str) and popt == '':
            bandgap = [np.nan, np.nan]
            try:
                bandgap = self.bandgapFromTauc(self.x(), self.y())
            except Exception:
                if not self.silent:
                    print('ERROR readDataFromFileEQE during Eg calculation',
                          self.filename, self.data[-1].data())
            # also useful to store information that we were *not* able to
            # compute bandgap
            self.update({'bandgap': bandgap[0], 'bandgapslope': bandgap[1]})
        # for saved files further re-reading
        self.update ({'Curve': CurveEQE.CURVE})



    # GUI RELATED FUNCTIONS
    def funcListGUI(self, **kwargs):
        out = Curve.funcListGUI(self, **kwargs)
        # format: [func, 'func label', ['input 1', 'input 2', 'input 3', ...]]
        # tauc (E*EQE)**2
        if (not self.attributeEqual('_popt')
            and self.attributeEqual('_fitfunc', 'func_bandgapTaucCurve')):
            out.append([self.updateFitParam, 'Update fit', ['Eg', 'slope'],
                        roundSignificant(self.getAttribute('_popt'),5)])
        elif self.attributeEqual('_popt'):
            mulOffset = self.getAttribute('mulOffset', default=1)
            out.append([self.CurveEQE_bandgap_print,
                        'Bandgap fit Tauc (eV*EQE)^2', ['Range EQE'],
                       ['['+str(0.25*mulOffset)+', '+str(0.70*mulOffset)+']']])
        # Eg from derivative method
        if (not self.attributeEqual('_popt')
            and self.attributeEqual('_fitfunc', 'bandgapDeriv_savgol')):
            out.append([None, 'Cannot update, must create a new one from '
                        +'initial EQE curve', [], []])
        elif self.attributeEqual('_popt'):
            out.append([self.CurveEQE_bandgapDeriv_print, 'Bandgap derivative',
                        ['Savitsky Golay width', 'degree'], [5,2]])
        # EQE current
        if self.attributeEqual('_popt'):
            out.append([self.currentCalc, 'EQE current', ['ROI', 'interpolate', 'spectrum'],
                        [[min(self.x()), max(self.x())], 'cubic', self.EQE_AM15_REFERENCE.replace('.txt','')], {},
                        [{'width': 15},
                         {'width': 8, 'field':'Combobox', 'values':['linear', 'quadratic', 'cubic']},
                         {'width': 8, 'field':'Combobox', 'values':[self.EQE_AM15_REFERENCE.replace('.txt',''), self.EQE_AM0_REFERENCE.replace('.txt','')]}]])
        # EQE 20.4%
        out.append([self.CurveEQE_Empa20p4, self.EQE_BEST_CELL_LABEL, [], []])
        # tauc (E*ln(1-EQE))**2
        if (not self.attributeEqual('_popt')
            and self.attributeEqual('_fitfunc', 'func_bandgapTaucCurveLog')):
            out.append([self.updateFitParam, 'Update fit', ['Eg', 'slope'],
                        roundSignificant(self.getAttribute('_popt'),5)])
        elif self.attributeEqual('_popt'):
            mulOffset = self.getAttribute('mulOffset', default=1)
            out.append([self.CurveEQE_bandgapLog_print,
                        'Bandgap fit Tauc (eV*ln(1-EQE))^2', ['Range EQE'],
                    ['['+str(0.30*mulOffset)+', '+str(0.88*mulOffset)+']']])
        # Urbach fitting
        if self.attributeEqual('_popt'):
            try: # Urbach fitting
                from grapa.datatypes.curveArrhenius import CurveArrheniusExpDecay
                try:
                    x0 = 1 + np.where(list(reversed(self.y() > 0.1)))[0][0]
                except IndexError:
                    x0 = 1
                try:
                    x1 = 1 + np.where(list(reversed(self.y() > 8e-4)))[0][0]
                except IndexError:
                    x1 = len(self.x())
                x0, x1 = self.shape()[1] - x0, self.shape()[1] - x1
                ROI = [self.x(index=x0, alter='nmeV'), self.x(index=x1, alter='nmeV')]
                ROI.sort()
                mulOffset = self.getAttribute('mulOffset', 1)
                out.append([self.CurveUrbachEnergy,
                            CurveArrheniusExpDecay.BUTTONFitLabel(),
                            [CurveArrheniusExpDecay.BUTTONFitROI()], [ROI]])
            except ImportError as e:
                print('WARNING CurveEQE: do not find grapa.datatypes.curveArrhenius.')
                print(type(e), e)
            
        out.append([self.printHelp, 'Help!', [], []])
        return out
    
    def alterListGUI(self):
        out = Curve.alterListGUI(self)
        out.append(['nm <-> eV', ['nmeV', ''], ''])
        out.append(['Tauc plot', ['nmeV', 'tauc'], ''])
        if (not self.attributeEqual('_popt')
            and self.attributeEqual('_fitfunc', 'func_bandgapTaucCurveLog')):
            out.append(['Tauc plot (E ln(1-EQE)^2)', 
                        ['nmeV', 'taucln1-eqe'], ''])
        return out



    # FUNCTIONS RELATED TO FIT

    def bandgapFromTauc(self, nm, EQE, yLim=[.25, .70], xLim=[600, 1500],
                        mode='EQE') :
        """
        Performs fit of low-energy side of EQE and returns [bandgap, slope].
        Executed at initialization.
        mode: 'EQE', or 'ln1-EQE'
        """
        if max(EQE) > 10 :
            print ('Function bandgapFromTauc: max(EQE) > 1. Multiplied by 0.01'
                   + ' for datapoint selection.')
            EQE = EQE * 0.01
        # select suitable data range
        mask = np.ones(len(nm), dtype=bool)
        for i in reversed(range (len (mask))) :
            if (EQE[i] < yLim[0] or EQE[i] > yLim[1] or nm[i] < xLim[0]
                or nm[i] > xLim[1]):
                mask[i] = False
        nm = nm[mask]
        EQE = EQE[mask]
        # perform fit
        eV = Curve.NMTOEV / nm
        tauc = (eV * np.log(1-EQE))**2 if mode == 'log1-EQE' else (eV * EQE)**2
        if len (tauc > 1):
            z = np.polyfit(eV, tauc, 1, full=True)[0]
            bandgap = -z[1]/z[0] #p = np.poly1d(z)
            return [bandgap, z[0]]
        #print ('Function bandgapFromTauc: not enough suitable datapoints.')
        return [np.nan, np.nan]
    
    def CurveEQE_bandgap_print(self, yLim=None):
        out = self.CurveEQE_bandgap(yLim=yLim)
        print ('Bandgap (Tauc (E*EQE)^2):',
               roundSignificant(out.getAttribute('_popt')[0],4), 'eV')
        return out
    def CurveEQE_bandgap(self, yLim=None):
        if yLim is None:
            yLim = [.25, .70]
        try:
            yLim = np.array(yLim) / self.getAttribute('mulOffset', default=1)
            bandgap = self.bandgapFromTauc(self.x(), self.y(), yLim=yLim)
        except:
            pass
        x = self.x()
        nm = np.arange(min(x), max(x), (max(x)-min(x))/1000)
        fit = self.func_bandgapTaucCurve(nm, *bandgap)
        return CurveEQE([nm, fit],
                {'_popt': bandgap, '_fitFunc': 'func_bandgapTaucCurve',
                 'color': 'k', 'data filename': self.getAttribute('filename'), 
                 'muloffset': self.getAttribute('mulOffset', default=1)})
    def func_bandgapTaucCurve(self, nm, *bandgap):
        z = [bandgap[1], -bandgap[0]*bandgap[1]]
        p = np.poly1d(z)
        fit = p(Curve.NMTOEV / nm)
        for i in range(len(fit)):
            if fit[i] >=0:
                fit[i] = np.sqrt(fit[i]) / (Curve.NMTOEV / nm[i])
            else:
                fit[i] = np.nan
        return fit
  
    def CurveEQE_bandgapLog_print(self, yLim=None):
        out = self.CurveEQE_bandgapLog(yLim=yLim)
        print ('Bandgap (Tauc (E*ln(1-EQE))^2):',
               roundSignificant(out.getAttribute('_popt')[0],4), 'eV')
        return out
    def CurveEQE_bandgapLog(self, yLim=None):
        if yLim is None:
            yLim = [.25, .70]
        try:
            yLim = np.array(yLim) / self.getAttribute('mulOffset', default=1)
            bandgap = self.bandgapFromTauc(self.x(), self.y(), yLim=yLim,
                                           mode='log1-EQE')
        except:
            pass
        x = self.x()
        nm = np.arange(min(x), max(x), (max(x)-min(x))/1000)
        fit = self.func_bandgapTaucCurveLog(nm, *bandgap)
        return CurveEQE([nm, fit],
                        {'_popt': bandgap,
                        '_fitFunc': 'func_bandgapTaucCurveLog', 'color': 'k',
                        'data filename': self.getAttribute('filename'),
                       'muloffset': self.getAttribute('mulOffset', default=1)})
    def func_bandgapTaucCurveLog(self, nm, *bandgap):
        z = [bandgap[1], -bandgap[0]*bandgap[1]]
        p = np.poly1d(z)
        fit = p(Curve.NMTOEV / nm)
        for i in range(len(fit)):
            if fit[i] >=0:
                fit[i] = 1 - np.exp(- np.sqrt(fit[i]) / (Curve.NMTOEV/nm[i]))
            else:
                np.nan
        return fit


    def bandgapDeriv(self, SGwidth=None, SGdegree=None, x_y_savgol=None):
        """
        Computes the bandgap based on the savitsky golay derivative,
        with a gaussian fit to the highest values
        SGwidth: 5. Parameters of Savitsky Golay filtering
        SGdegree: 3
        x_y_savgol (optional) to avoid calculating twice if already computed
            somewhere else
        """
        # need dummy x argument to use the updateFitParam method
        from scipy.optimize import curve_fit
        if x_y_savgol is None:
            x_, y_, savgol = self.bandgapDeriv_savgol(SGwidth, SGdegree)
        else:
            x_, y_, savgol = x_y_savgol
        order = True if x_[1] > x_[0] else False
        # locate max, fit gaussian around location of max
        iMax_ = np.argmax(savgol)
        ROIfit = range(max(0, iMax_-1 - int(not order)),
                       min(len(savgol), iMax_+2 + int(order)))
        datax, datay = x_[ROIfit], savgol[ROIfit]
        p0 = [savgol[iMax_], x_[iMax_], 2*(datax[1]-datax[0])]
        def func_gaussSimple(x, a, x0, sigma):
            return a * np.exp( - (x-x0)**2 / (2*sigma**2))
        popt, pcov = curve_fit(func_gaussSimple, datax, datay, p0=p0)
        self.update({'bandgapDeriv': Curve.NMTOEV/popt[1]})
        return popt
    def bandgapDeriv_savgol(self, SGwidth=None, SGdegree=None):
        """ returns a reduced x,y range and the savitsky golay derivative """
        from scipy.signal import savgol_filter
        if SGwidth is None:
            SGwidth = 5
        if SGdegree is None:
            SGdegree = 2
        SGdegree = int(SGdegree)
        SGwidth = int(SGwidth)
        if SGdegree >= SGwidth:
            print('Curve EQE bandgap derivative: degree must be lower than width for Savitsky-Golay filtering.')
            return None, None, None
        if SGwidth % 2 == 0:
            print('Curve EQE bandgap derivative: width must be odd for Savitsky-Golay filtering.')
            return None, None, None
        #compute derivative
        x, y = self.x(), self.y()
        if len(x) < 5:
            print('Not enough datapoints, cannot continue with processing.')
            return False
        # determines if sorted asc. or desc. Do NOT check if not monotoneous
        order = True if x[1] > x[0] else False
        # data selection - up to EQE maximum
        iMax = np.argmax(y)
        if order:
            ROI = range(max(0,iMax-2), len(y))
        else:
            ROI = range(0, min(iMax+2, len(y)))
        x_, y_, eV = x[ROI], y[ROI], Curve.NMTOEV / x[ROI]
        # locate peak using Savitsky Golay filtering; then correct for possible
        # uneven data point spacing
        savgol = savgol_filter(y_, SGwidth, SGdegree, deriv=1)
        savgol /= np.append(np.append(eV[1]-eV[0], (eV[2:]-eV[:-2])/2),
                                      eV[-1]-eV[-2])
        return x_, y_, savgol
    def CurveEQE_bandgapDeriv_print(self, SGwidth=None, SGdegree=None):
        out = self.CurveEQE_bandgapDeriv(SGwidth=SGwidth, SGdegree=SGdegree)
        print ('Bandgap (derivative):',
               roundSignificant(out.getAttribute('bandgapDeriv'), 4), 'eV')
        return out
    def CurveEQE_bandgapDeriv(self, SGwidth=None, SGdegree=None):
        """ returns the curve, with correspoding bandgap in the attributes """
        x_, y_, savgol = self.bandgapDeriv_savgol(SGwidth=SGwidth,
                                                  SGdegree=SGdegree)
        popt = self.bandgapDeriv(SGwidth=SGwidth, SGdegree=SGdegree,
                                 x_y_savgol=[x_,y_,savgol])
        out = CurveEQE([x_, savgol],
                       {'color': 'k', '_popt': [SGwidth, SGdegree],
                       'bandgapDeriv': Curve.NMTOEV/popt[1],
                       '_fitFunc': 'bandgapDeriv_savgol',
                       'data filename': self.getAttribute('filename')})
        out.update({'muloffset': -self.getAttribute('mulOffset', 1)*0.1,
                    'offset':-max(out.y())})
        return out




    def CurveEQE_Empa20p4(self):
        """ Returns the 20.4% Empa cell EQE. """
        import os.path
        from grapa.graph import Graph
        path = os.path.dirname(os.path.abspath(__file__))
        graph = Graph(os.path.join(path, self.EQE_BEST_CELL_REF), silent=True)
        if graph.length() > 0:
            return graph.curve(-1)
        graph = Graph(self.EQE_BEST_CELL_REF, silent=True)
        if graph.length() > 0:
            return graph.curve(-1)
        graph = Graph('_modules/'+self.EQE_BEST_CELL_REF, silent=True)
        if graph.length() > 0:
            return graph.curve(-1)
        print ('Data file EQE Empa 20.4% not found!')



    # other functions
    def CurveEQE_returnAM15referenceCurve(self):
        """ returns AM1.5 reference spectrum """
        import os.path
        from grapa.graph import Graph
        path = os.path.dirname(os.path.abspath(__file__))
        ref = Graph(os.path.join(path, self.EQE_AM15_REFERENCE), silent=True)
        if ref.length() > 0:
            return ref.curve(1)
        ref = Graph(self.EQE_AM15_REFERENCE, silent=True)
        if ref.length() > 0:
            return ref.curve(1)
        ref = Graph('_modules/'+self.EQE_AM15_REFERENCE, silent=True)
        if ref.length() > 0:
            return ref.curve(1)
        print('ERROR CurveEQE_returnAM15reference: cannot find reference file',
              self.EQE_AM15_REFERENCE)
        return 0
    def CurveEQE_returnAM0referenceCurve(self):
        """
        Returns AM0 reference spectrum.
        Data are in W/m2/um, returns spectral photon irradiance
        """
        from grapa.graph import Graph
        path = os.path.dirname(os.path.abspath(__file__))
        ref = Graph(os.path.join(path, self.EQE_AM0_REFERENCE))
        if ref.length() > 0:
            # h * c / wavelength -> energy WL conversion in vacuum for AM0
            energy = 6.62607004E-34 * 299792458 / (1e-9 * ref.curve(0).x())
            irrad = ref.curve(0).y() / energy / 1000
            return Curve([ref.curve(0).x(), irrad], {})
        return False
    
    # other functions
    def _curveReference(self, file):
        """ returns some reference spectrum """
        if file.endswith('.txt'):
            file = file[:-4]
        if file == 'AM1-5_Ed2-2008':
            return self.CurveEQE_returnAM15referenceCurve()
        if file == 'AM0_2000_ASTM_E-490-00':
            return self.CurveEQE_returnAM0referenceCurve()
        import os.path
        from grapa.graph import Graph
        path = os.path.dirname(os.path.abspath(__file__))
        ref = Graph(os.path.join(path, file), silent=True)
        if ref.length() > 0:
            return ref.curve(0)
        print('ERROR CurveEQE._curveReference: cannot find reference file', file)
        return False
    
    
    def currentCalc(self, ROI=None, interpolatekind='cubic',
                    spectralPhotonIrrad=None, silent=False):
        """
        Computes the current EQE current using AM1.5 spectrum
        ROI: [nm_min, nm_max]
        interpolatekind: order for interpolation of EQE data. default 'cubic'.
        refSpectrum: by default AM1.5G; otherwise filename in folder datatypes
        """
        if spectralPhotonIrrad is None:
            spectralPhotonIrrad = CurveEQE.EQE_AM15_REFERENCE
        refIrrad = self._curveReference(spectralPhotonIrrad)
        if not refIrrad:
            print('Error CurveEQE currentCalc, cannot find reference spectrum',
                  spectralPhotonIrrad)
            return
        ROIdef = [min(self.x()), max(self.x())]
        if ROI is None:
            ROI = ROIdef
        else:
            ROI = [max(ROIdef[0], ROI[0]), min(ROIdef[1], ROI[1])]
        # localize range of interest
        refROI = (refIrrad.x() >= ROI[0]) * (refIrrad.x() <= ROI[1])
        refDataX = refIrrad.x()[refROI]
        refDataY = refIrrad.y()[refROI]
        # interpolate data on ref x sampling
        # spline interpolation degree 1
        f = interpolate.interp1d(self.x(), self.y(), kind=interpolatekind)
        interpData = f(refDataX)
        # compute final spectrum
        finalSpectrum = refDataY * interpData
        # integriere QE*spektrum (auf gewähltem range)
        Jo = np.trapz(finalSpectrum, refDataX)
        EQEcurrent = Jo*1.602176487E-19/10
        if not silent:
            print('Curve', self.getAttribute('label'), 'EQE current:', 
                  EQEcurrent, 'mA/cm2')
        return EQEcurrent
        
    
    def CurveUrbachEnergy(self, ROIeV):
        """ Return fit in a Curve object, giving Urbach energy decay """
        from datatypes.curveArrhenius import CurveArrhenius, CurveArrheniusExpDecay
        ROIeV.sort()
        curve = CurveArrhenius(self.data, CurveArrheniusExpDecay.attr)
        out = curve.CurveArrhenius_fit(Tlim=ROIeV)
        out.update(self.getAttributes(['offset', 'muloffset']))
        return out
        
    
    def printHelp(self):
        print('*** *** ***')
        print('CurveEQE offer basic treatment of (external) quantum')
        print('efficiency of solar cells.')
        print('Input units should be [nm] and [0-1] (if needed, CurveSpectrum')
        print('can convert [eV] in [nm]).')
        print('Curve transforms:')
        print(' - nm <-> eV: switches the horizontal axis from nm to eV.')
        print(' - Tauc plot: displays (eV*EQE)^2 vs eV. The corresponding')
        print('   Tauc fit is a straight line.')
        print(' - (optional) Tauc plot (E ln(1-EQE)^2): only appears if a ')
        print( '  Curve was fitted with the corresponding formula.')
        print('Analysis functions:')
        print(' - Bandgap fit Tauc (eV*EQE)^2: fit the bandgap with the')
        print(' indicated formula. Parameters:')
        print('   Range EQE: select data with EQE value in indicated range.')
        print(' - Bandgap derivative: determines the bandgap following the')
        print('   derivative approach. The derivative dEQE/dE is used instead')
        print('   of dEQE/dlambda as the resulting curve is often more')
        print('   symmetric. The algorithm computes the derivative with the')
        print('   Savitsky-Golay algorithm. Parameters:')
        print('   Savitsky Golay width, and degree: width should be odd, and')
        print('      degree must be between 1 and width. By default 5, 2.')
        print('      With parameters 3, 1 ones retrieves the symmetrical')
        print('      discrete difference (y_1+1 - y_i-1)  / (x_i+1 - x_i-1).')
        print(' - EQE current: integration of the product of the EQE with the')
        print('   reference AM1.5 solar spectrum '+self.EQE_AM15_REFERENCE+'.')
        print('   The code is similar to that of CurrentCalc matlab code.')
        print(' - Bandgap fit Tauc (eV*ln(1-EQE))^2: fits the bandgap with')
        print('   the indicated formula. In principle more exact, however')
        print('   highly sensitive to reflective and collection losses. Less')
        print('   robust than (eV*EQE)^2. Parameters:')
        print('   Range EQE: select data with EQE value in indicated range.')
        print(' - Urbach energies from fit to exponential decay.')
        print('   A new Curve is created, with main parameters the decay')
        print('   energy U and the energy at 100% EQE.')
        print('   ROI [eV]: limits to the fitted data, in eV.')
        return True

        


# -*- coding: utf-8 -*-
"""
Created on Fri Jul  1 15:22:51 2016

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import numpy as np


from grapa.graph import Graph
from grapa.mathModule import is_number


class Database:

    def __init__(self, data, colLabels=[], rowLabels=[], rowLabelNums=[]):
        """ can be instancied 2 ways:
        - by providing a Graph object
        - by provinding data: np.array, colLabels, rowLabels the labels of the columns, respetively the rows, and rowLabelNums a numeric index corresponding the row label.
        """
        self.rowLabel = 'Growth TL nb'
        self.attr = {}
        if isinstance(data, Graph):
            self.colLabels = data.getAttribute('collabels')
            self.rowLabels = data.getAttribute('rowlabels')
            self.rowLabelNums = data.getAttribute('rowlabelnums')
            if data.getAttribute('rowLabel') != '':
                self.rowLabel = data.getAttribute('rowLabel')
            if data.curve(0).shape(1) == 0:
                self.data = np.ndarray([])
            else:
                # assumes all xyCurves have identical x data
                self.data = np.ndarray((data.curve(0).shape(1), data.length()))
                for i in range(data.length()):
                    self.data[:, i] = data.curve(i).y()
                    if not np.array_equiv(data.curve(i).x(), data.curve(0).x()):
                        print('ERROR database init: columns with different rows x values! (col', i, ')')
#                        for j in range(len(data.curve(i).x())):
#                            if data.curve(i).x(j) != data.curve(0).x(j):
#                                print ('   ', j, data.curve(i).x(j), data.curve(0).x(j))
        else:
            self.data = np.array(data)
            self.colLabels = colLabels
            self.rowLabels = rowLabels
            if rowLabelNums == []:
                self.rowLabelNums = self.rowLabels[:]
            else:
                self.rowLabelNums = rowLabelNums
            for i in range(len(self.rowLabelNums)):
                try:
                    self.rowLabelNums[i] = float(self.rowLabelNums[i])
                except Exception:
                    self.rowLabelNums[i] = np.nan

    def update(self, attr):
        for key in attr :
            self.attr.update({key.lower(): attr[key]})
    def getAttr(self, key):
        if key.lower() in self.attr:
            return self.attr[key.lower()]
        return ''

    def shape(self, idx=''):
        """ Returns the shape of the np.ndarray data container. """
        if idx == '':
            return self.data.shape
        return self.data.shape[idx]

    def getColLabel(self, i=':'):
        """ Returns label of column i. """
        if i == ':':
            return self.colLabels
        return self.colLabels[i]

    def getRowLabel(self, i=':'):
        """ Return the row label. """
        if i == ':':
            return self.rowLabels
        return self.rowLabels[i]

    def getRowLabelNum(self, i=':'):
        """ Return the numeric equivalent for the row label. """
        if i == ':':
            return self.rowLabelNums
        return self.rowLabelNums[i]

    def renameColumn(self, oldname, newname):
        """ Change the label of a column. """
        if is_number(oldname):
            self.colLabels[oldname] = newname
        elif oldname in self.colLabels:
            self.colLabels[self.colLabels.index(oldname)] = newname
        else:
            print('WARNING database renameColumn: column', newname, ' not found.')
            print('   ', self.colLabels)

    def setRowLabel(self, newlabel, i=-1):
        """ Change the row label. If i=-1 self.rowLabel is modified (default 'Growth TL nb')  """
# change numeric rowLabelNum as well ?
        if i == -1:
            self.rowLabel = newlabel
        else:
            self.rowLabels[i] = newlabel



    def deleteRow(self, row_idx):
        self.rowLabels[row_idx] = ''
        for i in range(self.data.shape[1]):
            self.data[row_idx, i] = np.nan



    def merge(self, other):
        # merge rows
        le0 = self.data.shape[0]
        le1 = self.data.shape[1]
        le_ = other.data.shape
        if len(le_) == 1 and len(other.colLabels) == 1: # if only 1 column
            le_ = le_ + (1,)
        if len(le_) == 1 and len(other.colLabels) > 1: # if no entry
            le_ = le_ + (len(other.colLabels),)
        newdata = np.zeros((le0 + le_[0], le1 + le_[1]))
        newdata[0:le0, 0:le1] = self.data

        for i in range(other.data.shape[0]):
            try:
                idx = self.rowLabelNums.index(other.rowLabelNums[i])
                newdata = np.delete(newdata, -1, axis=0)
            except ValueError:
                idx = le0
                print('database merge: new line for index', i, '(', other.rowLabels[i], ')')
                self.rowLabels.append(other.rowLabels[i])
                self.rowLabelNums.append(other.rowLabelNums[i])
                le0 += 1
            if len(other.data.shape) == 1:
                newdata[idx, le1:] = other.data[i]
            else:
                newdata[idx, le1:] = other.data[i, :]
        self.data = newdata
        for col in other.colLabels:
            self.colLabels.append(col)





    def clean(self):
        for i in range(self.data.shape[0]):
            if (not self.data[i, :].any()):
                if self.rowLabels[i] == '':
                    print('Delete index ', i, self.rowLabels[i])
                    self.data = np.delete(self.data, i, axis=0)
                    del self.rowLabels[i]
                    del self.rowLabelNums[i]

    def rowIdFromLabel(self, y):
        return self.rowLabels.index(y)

    def colIdFromLabel(self, y):
        return self.colLabels.index(y)

    def value(self, col, row, silent=False):
        i = row if isinstance(row, int) else self.rowIdFromLabel(row)
        [val, collabel] = self.colValuesFromId(col)
        try:
            return val[i]
        except Exception:
            if not silent:
                print('Database value: cannot find col, row:', col, row, i, val)
        return np.nan

    def setValue(self, col, row, value, silent=False):
        i = row if isinstance(row, int) else self.rowIdFromLabel(row)
        if not isinstance(col, int):
            col = self.colIdFromLabel(col)
        try:
            self.data[i, col] = value
            return 0
        except Exception:
            if not silent:
                print('Database setValue: cannot find col, row:', col, row, i, value)
        return 1


    def colValuesFromId(self, x):
        xlabel = self.rowLabel
        if isinstance(x, int):
            xlabel = self.colLabels[x]
            x = self.data[:, x]
        else:
            if x.lower() in ['rowlabelnums', 'row']:
                x = self.rowLabelNums
            elif not is_number(x) and x in self.colLabels:
                xlabel = x
#                print('colValuesFromId', x, self.colLabels.index(x), self.data.shape, self.colLabels)
                x = self.data[:, self.colLabels.index(x)]
            else:
                if not is_number(x):
                    print('ERROR colValuesFromId', x, self.colLabels)
                xlabel = self.colLabels[x]
                x = self.data[:, x]
        return [np.array(x), xlabel]




    def maskFromConditions(self, tests):
        # tests can be [['Eff. [%]', '>=', 19], ['row', '<',  2700]]],
        # or [['VOC deficit', '<', 0.420]]
        # can handle several tests, but combination is always AND
        mask = list(np.ones(self.data.shape[0]))
        for test in tests:
            [vals, label] = self.colValuesFromId(test[0])
            for i in range(self.data.shape[0]):
                if test[1] == '==':
                    mask[i] = mask[i] * (vals[i] == test[2])
                if test[1] == '>':
                    mask[i] = mask[i] * (vals[i] >  test[2])
                if test[1] == '>=':
                    mask[i] = mask[i] * (vals[i] >= test[2])
                if test[1] == '<':
                    mask[i] = mask[i] * (vals[i] <  test[2])
                if test[1] == '<=':
                    mask[i] = mask[i] * (vals[i] <= test[2])
                if test[1] == '!=':
                    mask[i] = mask[i] * (vals[i] != test[2])
        mask2 = []
        for i in range(len(mask)):
            if mask[i] == 1:
                mask2.append(i)
        return mask2





    def plot2(self, x, y, linespec='x', xlim=[0, 0], ylim=[0, 0], complement={}, also=[], filesavesuffix=''):
        """ plot the data of
        Possible call:
            testEffId1 = [['Eff. [%]', '>=', 19], ['row', '<',  2700]]
            testEffId2 = [['Eff. [%]', '>=', 19], ['row', '>=', 2700], ['row', '<', 2884]]
            testEffId3 = [['Eff. [%]', '>=', 19], ['row', '>=', 2884]]
            def alsoEff(x, y, color):
                return [[x, y, 's'+color, testEffId1], [x, y, 'v'+color, testEffId2], [x, y, 'sr', testEffId3]]
            dbProcess.plot('row',          'Eff. [%]',     linespec='xk', xlim=[2350, np.ceil(max(dbProcess.rowLabelNums)/50)*50], ylim=[15, 21],  also=alsoEff('row',          'Eff. [%]', 'k'))
        """
        def cleanlabelfilename (label):
            return label.replace(' ', '').replace('[', '').replace(']', '').replace('%', '').replace('.', '')
            
        [x, xlabel] = self.colValuesFromId(x)
        [y, ylabel] = self.colValuesFromId(y)

        if 'linespec' not in complement:
            complement.update({'linespec': linespec})
        if 'xlabel' not in complement:
            complement.update({'xlabel': xlabel})
        if 'ylabel' not in complement:
            complement.update({'ylabel': ylabel})
        if xlim != [0, 0]:
            complement.update({'xlim': xlim})
        if ylim != [0, 0]:
            complement.update({'xlim': ylim})

        if filesavesuffix == '':
            filesavesuffix = 'dbPlot_'
        if filesavesuffix[-1] != '_':
            filesavesuffix += '_'
        filesave = './export/' + filesavesuffix + cleanlabelfilename(xlabel) + '_' + cleanlabelfilename(ylabel)
        graph = Graph([x, y], complement, silent=True)
        if also != []:
            #[ [x, y, linespec, [[colIdxCrit, crit, value], [colIdx, crit, value]] ] , second line, curve.]
            for curve in also:
                if curve[0] == '':
                    curve[0] = xlabel
                if curve[1] == '':
                    curve[1] = ylabel
                [x2, xlabel2] = self.colValuesFromId(curve[0])
                [y2, ylabel2] = self.colValuesFromId(curve[1])
                complement2 = {'linespec': curve[2]}
                mask = self.maskFromConditions(curve[3])
                if len(mask) > 0:
                    graph2 = Graph([x2[mask], y2[mask]], complement2, silent=True)
                    graph.merge(graph2)
                filesave += '_' + cleanlabelfilename(curve[3][0][0])

        if self.getAttr('savesilent') :
            graph.update({'savesilent': True})
        graph.plot(filesave=filesave)





    def plot(self, x, y, linespec='x', xlim=[0, 0], ylim=[0, 0], also=[]):
        """ deprecated plot function """
        self.plot2(x, y, complement={'linespec':linespec, 'xlim': xlim, 'ylim': ylim}, also=also)
#        [x, xLabel] = self.colValuesFromId(x)
#        [y, yLabel] = self.colValuesFromId(y)
#
#        fig,ax = plt.subplots()
#        fig.patch.set_alpha(0.0)
#
#        ax.plot(x, y, linespec)
#        ax.ticklabel_format(useOffset=False)
#        ax.set_ylabel(yLabel)
#        ax.set_xlabel(xLabel)
#        if ylim != [0,0]:
#            ax.set_ylim(bottom=ylim[0])
#            ax.set_ylim(top   =ylim[1])
#        if xlim != [0,0]:
#            ax.set_xlim(left= xlim[0])
#            ax.set_xlim(right=xlim[1])
#
#        if also != []:
#            #[ [x, y, linespec, [[colIdxCrit, crit, value], [colIdx, crit, value]] ] , second line, curve.]
#            for curve in also:
#                [x, xLabel] = self.colValuesFromId(curve[0])
#                [y, yLabel] = self.colValuesFromId(curve[1])
#                linespec = curve[2]
#                mask = self.maskFromConditions(curve[3])
#                if len(mask) > 0:
#                    ax.plot(x[mask], y[mask], linespec)
#                else:
#                    print('plot also: no data to plot')
#
#        filename = 'databasePlot_' + xLabel + '_' + yLabel
#        filesave = filename + '.png'
#        plt.savefig(filesave, transparent=True)
#
#        out = ''
#        out+= 'xlabel\t' + xLabel + '\n'
#        out+= 'ylabel\t' + yLabel + '\n'
#        out+= 'linespec\t' + linespec + '\n'
#        for i in range(len(x)):
#            out += str(x[i]) + '\t' + str(y[i]) + '\n'
#        f = open(filename + '.txt', 'w')
#        f.write(out)
#        f.close()
#



# -*- coding: utf-8 -*-
"""
@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import numpy as np
import copy


from grapa.curve import Curve
from grapa.mathModule import roundSignificant




def findXAtValuePolynom(datax, datay, target, ifPlot=False):
    if len(datax) < 3 or len(datay) < 3:
        print('ERROR CurveSIMS findXAtValuePolynom: not enough data to fit 3rd degree polynom, returned 0.')
        return [0, 0]
    z = np.polyfit(datax, datay, 3, full=True)[0]
    roots = np.roots((np.poly1d(z) - target))
    rootsReal = np.real(roots[np.isreal(roots)])
    rootsInROI = rootsReal[(rootsReal >= min(datax)) * (rootsReal <= max(datax))]
    if len(rootsInROI) == 0:
        rootsInROI = rootsReal
    if len(rootsInROI) == 0:
        print('Error findXAtValuePolynom cannot find suitable value.', target, datax, datay)
    root = np.average(rootsInROI)
    idx = np.argmin(np.abs(datax - rootsInROI))
    if ifPlot:
        import matplotlib.pyplot as plt
        plt.figure()
        plt.plot(datax, datay, '.-b')
        plt.plot(datax, np.polyval(z, datax), 'k')
    return [root, idx]
    
# identify transition from low to high level
def selectDataEdge(curve, ROI, targetRel=None, threshRel=None, ifTrailingEdge=False):
    if targetRel is None:
        targetRel = 0.5
    if threshRel is None:
        threshRel = [targetRel*0.6, targetRel*0.6+1*0.4]
    y = copy.deepcopy(curve.y(index=ROI))
    iMin, iMax = y.argmin(), y.argmax()
    targetAbs = y[iMin] + targetRel * (y[iMax] - y[iMin])
    # data normalization in [0,1] interval
    y = (y - y[iMin]) / y[iMax]
    if not ifTrailingEdge:
        rangeThres = [ROI[0] +        (y        > t).argmax() for t in threshRel]
    else:
        rangeThres = [ROI[0] + iMax + (y[iMax:] < t).argmax() for t in threshRel]
    rangeThres.sort()
    rangeThres = range(rangeThres[0], rangeThres[1]+1)
    datax = curve.x(index=rangeThres)
    datay = curve.y(index=rangeThres)
    idx = ROI[0] + iMax + (y[iMax:] < 0.5 * y[iMax]).argmax()
    return [datax, datay, targetAbs, rangeThres]




class CurveSIMS(Curve):
    
    CURVE = 'Curve SIMS'
    
    def __init__(self, *args, **opts):
        Curve.__init__(self, *args, **opts)
        self.update({'Curve': CurveSIMS.CURVE})
        # SIMS data files are a bit special, header start with empty cell
        if self.getAttribute('label') == '' and self.getAttribute('empty0') != '':
            self.update({'label': self.getAttribute('empty0'), 'empty0': ''})
        # some additional attributes, which must not override data saved in the data file
        add = {'_SIMSelement': self.getAttribute('label'), '_SIMSmsmt': ''.join(self.getAttribute('filename').split('/')[-1].split('\\')[-1].split('.')[:-1]).replace('_',''), '_SIMStotal': 0, '_SIMSYieldCoef': 1}
        for key in add:
            if self.getAttribute(key) == '':
                self.update({key: add[key]})
        # check and clean first 2 empty lines (total and header lines might have been mistaken)
        if np.isnan(self.x(index=1)) and np.isnan(self.y(index=1)):
            self.update({'_SIMStotal': self.y(index=0)})
            self.data = np.array([self.x()[2:], self.y()[2:]])

    # RELATED TO GUI
    def funcListGUI(self, **kwargs):
        from grapa.datatypes.graphSIMS import GraphSIMS
        msmtIdDict = {'msmtId': self.getAttribute('_SIMSmsmt')}
        ratios = [k for k in GraphSIMS.KEYWORDS.keys()] + ['other ratio']# + ['ratio as ["elem0"],["elem1","elem2"]']
        out = Curve.funcListGUI(self, **kwargs)
        # format: [func, 'func label', ['input 1', 'input 2', 'input 3', ...] (, [default1, default2, ...], {'hiddenVar1':value1}) ]
        out.append([self.localizeEdgeFromGUI, 
                    'Find edge',
                    ['ROI', 'target', 'threshold', 'edge'],
                    [[self.x(0), self.x(self.shape(1)-1)], 0.5, [0.3,0.7], 'leading'],
                    {},
                    [{},{},{},{'field':'Combobox', 'values':['leading','trailing'], 'width':7}]])
        ROI = [self.x(0), self.x(self.shape(1)-1)] if self.getAttribute('_SIMSLayerBoundaries') == '' else self.getAttribute('_SIMSLayerBoundaries')
        depth = (ROI[1] - ROI[0]) * self.getAttribute('_SIMSdepth_mult') if self.getAttribute('_SIMSdepth_mult') != '' else 2
        out.append([GraphSIMS.setLayerBoundariesDepthParametersGUI, 'Calibrate depth', ['Depth', 'ROI'], [depth, roundSignificant(ROI,5)], msmtIdDict])
        labels = ['^113In+']
        if 'graph' in kwargs:
            labels = []
            for c in kwargs['graph'].curves('_SIMSmsmt', msmtIdDict['msmtId']):
                labels.append(c.getAttribute('_SIMSelement'))
            if len(labels) == 0:
                labels = ['^113In+']
        out.append([GraphSIMS.targetRatioSetYieldGUI,
                    'Adjust compos.',
                    ['ROI', 'Ratio', 'Elem.', 'Compos.'],
                    [ROI, 'GGI', ('^113In+' if '^113In+' in labels else labels[0]), 0.35],
                    msmtIdDict,
                    [{}, {'field':'Combobox', 'values':ratios, 'width':6}, {'field':'Combobox', 'values':labels}, {}]])
        out.append([GraphSIMS.getRatioGUI, 'Compute ratio', ['ROI', 'Ratio'],
                    [ROI, 'GGI'], msmtIdDict,
                    [{}, {'field':'Combobox', 'values':ratios}]])
        out.append([GraphSIMS.appendReplaceCurveRatioGUISmt,
                    'Create curve ratio',
                    ['Ratio', 'Curve name', 'Smooth S-G w', 'd'],
                    ['GGI', 'GGI', 1, 1],
                    msmtIdDict,
                    [{'field':'Combobox', 'values':ratios, 'width':6}, {}, {}, {}]])
        out.append([self.printHelp, 'Help!', [], []])
        return out

        
    def alterListGUI(self):
        out = Curve.alterListGUI(self)
        out += [['Semilogy', ['', 'idle'], 'semilogy']]
        out += [['Raw <-> Depth', ['SIMSdepth', ''], 'semilogy']]
        return out
    # alter: s to z, normalized (in toggle switch?)


    def setDepthParameters(self, offset, mult):
        self.update({'_SIMSdepth_offset': offset, '_SIMSdepth_mult': mult})
    



    # other methods
    def localizeEdge(self, ROI, targetRel, threshRel=None, ifTrailingEdge=False, ROIrefinement=0):
        #☺ check if ROI is a actual range or only extrema
        if isinstance(ROI, (list, tuple)) and len(ROI) == 2 and np.abs(ROI[1]-ROI[0]) > 1:
            ROI.sort()
            ROI = range(int(ROI[0]), int(ROI[1]))
#            print ('Localize edge ROI correction (', min(ROI), max(ROI), ')')
        else:
            pass
        # first localize max in ROI, then compute max*target and select data subset [max*threshold[0], max*threshold[1]]
        [datax, datay, targetAbs, rangeThresholds] = selectDataEdge(self, ROI, targetRel, ifTrailingEdge=ifTrailingEdge, threshRel=threshRel)
        # then interpolate data subset with polynom, and find intersection with target*max
        [root, idx] = findXAtValuePolynom(datax, datay, targetAbs)
        idx += rangeThresholds[0]
        if ROIrefinement > 0:
            delta = ROIrefinement * (np.array([min(rangeThresholds), max(rangeThresholds)]) - idx)
            delta[0], delta[1] = idx + min(delta[0], idx-1), idx + max(delta[1], idx+1) # check at least 1 point enlargment
            delta[0], delta[1] = max(0, delta[0]), min(delta[1], len(self.y())-1) # check indices are valid
            delta[0], delta[1] = max(min(ROI), delta[0]), min(delta[1], max(ROI)) # do not want a larger ROI than before
            [root, idx] = self.localizeEdge(range(delta[0], delta[1]), targetRel, threshRel=threshRel, ifTrailingEdge=ifTrailingEdge)
        return [root, idx]
    
    def localizeEdgeFromGUI(self, ROI, targetRel, threshRel, ifTrailingEdge):
        # assumes ROI is in the form [sputterTimeMin, sputterTimeMax]
        if ifTrailingEdge == 'trailing':
            ifTrailingEdge = True
        elif ifTrailingEdge == 'leading':
            ifTrailingEdge = False
        if len(list(ROI)) > 2:
            ROI = [min(ROI), max(ROI)]
        ROI = [np.argmin(np.abs(self.x() - ROI[0])), np.argmin(np.abs(self.x() - ROI[1]))]
        res = self.localizeEdge(ROI, targetRel, threshRel=threshRel, ifTrailingEdge=ifTrailingEdge)
        return 'Edge position for curve ' + self.getAttribute('_SIMSelement') + ' at: ' + str(roundSignificant(res[0],5))
        

    # find indices of 
    def findLayerBorders(self, ROI=None, targetRel=0.5, threshRel=[0.3,0.7], ROIrefinement=0, returnIdx=True):
        # targetRel = None#0.5
        # threshRel = None # [0.3, 0.7]
        if ROI is None: # [range(0,10), range(10,20)]
            ROI = [range(0, int(self.shape(1)/2)), range(int(self.shape(1)/2), int(self.shape(1)))]
        posInterfGaIn = [np.nan, np.nan]
        for i in range(2):
            [root, idx] = self.localizeEdge(ROI[i], targetRel, threshRel=threshRel, ifTrailingEdge=i, ROIrefinement=ROIrefinement)
            posInterfGaIn[i] = idx if returnIdx else root
        return posInterfGaIn

        
    def printHelp(self):
        from grapa.datatypes.graphSIMS import GraphSIMS
        print('*** *** ***')
        print('CurveSIMS offer some basic processing of SIMS data.')
        print('Each can notably compute ratios of curves, for which a relative yield has to be provided.')
        print('Curves ratios are noted with keywords such as "ggi", or as "[[\'^71Ga+\'],[\'^71Ga+\', \'^113In+\']]". For 1st and 2nd list the curves multiplied by their yield are summed, and the ratio is finally computed. An empty top or bottom list is replaced by 1.')
        print('List of existing keywords:')
        for key in GraphSIMS.KEYWORDS:
            print('   ' + key + ':', str(GraphSIMS.KEYWORDS[key]))
        print('Curve transforms:')
        print(' - Semilogy, with etching time as default x coordinate,')
        print(' - Raw <-> Depth: depth as x axis, provided depth was calibrated (see analysis functions).')
        print('Analysis functions:')
        print(' - Find edge: auto-detect edges in the selected curve. Parameters:')
        print('   ROI: range of interest, in units of input x data,')
        print('   Target: default 0.5,')
        print('   Threshold: default [0.3,0.7]')
        print('   ifTrailing: 0 if leading edge, 1 if trailing edge.')
        print(' - Calibrate depth: calibrate depth by linear scaling of input x axis. Parameters:')
        print('   Depth: know depth of ROI,')
        print('   ROI: range of interest, in units of input x data.')
        print(' - Adjust composition: adjust the relative SIMS yield of a curve such that the chosen')
        print('   curve ratio has the chosen value within the selected range of interest. Parameters:')
        print('   ROI: range of interest, in units of input x data,')
        print('   Ratio: which Curve ratio is to be adjusted,')
        print('   Element: which element the yield has to be adjusted,')
        print('   Compos: the know composition we want to reach by adjusting the element yield.')
        print(' - Compute ratio: computes a ratio of curves within a ROI.')
        print('   ROI: range of interest, in units of input x data,')
        print('   Ratio: which Curve ratio is to be computed.')
        print(' - Create curve ratio: creates a Curve computed as a ratio of Curves.')
        print('   Ratio: which Curve ratio is to be computed,')
        print('   Curve name: give it a beautiful name.')
        print('   Smooth S-G w, d: a Savitsky-Golay smoothening can be applied to the elemental traces')
        print('   prior to combination, using window and degree. No effect if w 1 d 1, possibly w 21 d 3.')
        print('   ')
        return True



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 23 20:37:38 2017

@author: Romain Carron
Copyright (c) 2018, Empa, Laboratory for Thin Films and Photovoltaics, Romain Carron
"""

import numpy as np
from re import findall as refindall

from grapa.graph import Graph
from grapa.graphIO import GraphIO
from grapa.curve import Curve
from grapa.mathModule import is_number, derivative



class GraphCf(Graph):

    FILEIO_GRAPHTYPE = 'C-f curve'
    
    AXISLABELS = [['Frequency', 'f', 'Hz'], ['Capacitance', 'C', 'nF']] 


    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', **kwargs):
        if fileExt == '.txt' and line1.strip()[:13] == 'Sample name: ' and fileName[0:4] == 'C-f_':
            return True
        return False
        
        
    def readDataFromFile(self, attributes, **kwargs):
        len0 = self.length()
        GraphIO.readDataFromFileGeneric(self, attributes)
        self.castCurve('Curve Cf', len0, silentSuccess=True)
        # label based on file name, maybe want to base it on file content
        self.curve(len0).update({'label': self.curve(len0).getAttribute('label').replace('C-f ','').replace(' Cp','').replace(' C','').replace('T=','').replace(' [nF]','').replace(' (nF)','')})
        # graph cosmetics
        self.update({'typeplot': 'semilogx', 'alter': ['', 'idle']})
        self.update({'xlabel': self.formatAxisLabel(GraphCf.AXISLABELS[0]),
                     'ylabel': self.formatAxisLabel(GraphCf.AXISLABELS[1])}) # default
        colName = self.curve(len0).getAttribute('frequency [Hz]', -1) # retrieve actual unit
        if isinstance(colName, str): # retrieve yaxis information in file, interpret it
            ylabel = colName.replace('C ', 'Capacitance ').replace('Cp ', 'Capacitance ').replace('(','[').replace(')',']')
            expr = '^(.* )\[(.*)\]$'
            f = refindall(expr, ylabel)
            if isinstance(f, list) and len(f) == 1 and len(f[0]) == 2:
                ylabel = [f[0][0], GraphCf.AXISLABELS[1][1], f[0][1]]
            self.update({'ylabel': self.formatAxisLabel(ylabel)}) # default
        # generate additional curves if interesting
        nb_add = 0
        # if phase is required
        if self.getAttribute('_CfLoadPhase', False) is not False:
            f = self.curve(len0).x()
            C = self.curve(len0).y() * 1e-9 # in [F], not [nF]
            conductance = None
            for c in range(len0+1, self.length()): # retrieve resistance curve, not always same place
                lbl = self.curve(c).getAttribute('frequency [hz]')
                if isinstance(lbl, str) and lbl[:1] == 'R':
                    conductance = 1 / self.curve(c).y()
                    break
            if conductance is None:
                print('ERROR Read', self.filename, 'as CurveCf with phase: cannot find R.')
            else:
                phase_angle = np.arctan(f * 2 * np.pi * C / conductance) * 180. / np.pi
                self.append(Curve([f, phase_angle], self.curve(len0).getAttributes()))
                nb_add += 1
            
        # delete Rp and Temperature
        for c in range(self.length()-1-nb_add, len0, -1):
            self.deleteCurve(c)
        # normalize with area C, Nyquist
        area = self.curve(len0).getAttribute('cell area (cm2)', None)
        if area is None:
            area = self.curve(len0).getAttribute('cell area', None)
        if area is None:
            area = self.curve(len0).getAttribute('area', None)
        if area is not None:
            self.curve(len0).setY(self.curve(len0).y() / area)
#            self.curve(len0+1).setX(self.curve(len0+1).x() / area) # uncomment if created Nyquist
#            self.curve(len0+1).setY(self.curve(len0+1).y() / area) # uncomment if created Nyquist
            self.curve(len0).update({'cell area (cm2)': area})
            self.update({'ylabel': self.getAttribute('ylabel').replace('F', 'F cm$^{-2}$')})
            if not self.silent:
                print('Capacitance normalized to area', self.curve(len0).getArea(),'cm2.')
 


class CurveCf(Curve):
    
    CURVE = 'Curve Cf'
    
    def __init__(self, data, attributes, silent=False):
        """
        Constructor with minimal structure: Curve.__init__, and set the
        'Curve' parameter.
        """
        Curve.__init__(self, data, attributes, silent=silent)
        self.update ({'Curve': CurveCf.CURVE})


    # GUI RELATED FUNCTIONS
    def funcListGUI(self, **kwargs):
        out = Curve.funcListGUI(self, **kwargs)
        # format: [func, 'func label', ['input 1', 'input 2', 'input 3', ...]]
        out.append([self.setArea, 'Set cell area', ['New area'], [self.getArea()]])
        out.append([self.printHelp, 'Help!', [], []])
        return out
    
    def alterListGUI(self):
        out = Curve.alterListGUI(self)
        #out.append(['nm <-> eV', ['nmeV', ''], ''])
        out.append(['semilogx', ['','idle'], 'semilogx'])
        out.append(['-dC / dln(f)', ['', 'CurveCf.y_mdCdlnf'], 'semilogx'])
        return out

    def y_mdCdlnf(self, index=np.nan, xyValue=None):
        # if ylim is set, keep the ylim restriction and not do not ignore ylim
        if xyValue is not None:
            return xyValue[1]
        # do not want to use self.x(index) syntax as we need more points to 
        # compute the local derivative. HEre we compute over all data, then
        # restrict to the desired datapoint
        val = - derivative(np.log(self.x(xyValue=xyValue)), self.y(xyValue=xyValue))
        if len(val) > 0:
            if np.isnan(index).any():
                return val[:]
            return val[index]
        return val
    
    # FUNCTIONS RELATED TO GUI (fits, etc.)
    def setArea(self, value):
        oldArea = self.getArea()
        self.update({'cell area (cm2)': value})
        self.setY(self.y() / value * oldArea)
        return True
    def getArea(self):
        return self.getAttribute('cell area (cm2)', 1)
        

    # Function related to data picker
    def getDataCustomPickerXY(self, idx, **kwargs):
        """
        Overrides Curve.getDataCustomPickerXY
        Returns Temperature, omega instead of f, C
        """
        if 'strDescription' in kwargs and kwargs['strDescription']:
            return '(Temperature, omega) instead of (f, C)'
        # actually fo things
        try:
            from grapa.datatypes.curveArrhenius import CurveArrheniusCfdefault
            attr = CurveArrheniusCfdefault.attr
        except Exception:
            attr = {'Curve': 'Fit Arrhenius',
                    '_Arrhenius_variant': 'Cfdefects',
                    'label': 'omega vs Temperature'}
            print('WARNING Exception during opening of CurveArrhenius module. Does not perturb saving of the data.')
        attr.update({'sample name': self.getAttribute('sample name')})
        T = np.nan
        for key in ['temperature [k]', 'temperature']:
            T = self.getAttribute(key, np.nan)
            if not np.isnan(T):
                if not is_number(T):
                    T = float(T)
                break
        if not np.isnan(T):
            print('Data picker CurveCf T =',T,', omega =',self.x(idx)[0]*2*np.pi,'.')
            return T, self.x(idx)[0] * 2 * np.pi, attr
        return Curve.getDataCustomPickerXY(self, idx, **kwargs)
        
        
    def printHelp(self):
        print('*** *** ***')
        print('CurveCf offer basic treatment of C-f curves of solar cells.')
        print('Default units are frequency [Hz] and [nF] (or [nF cm-2]).')
        print('Curve transforms:')
        print(' - Linear: standard is Capacitance per area [nF cm-2] vs [Hz].')
        print(' - semilogx: horizontal axis is log(f) instead of f.')
        print(' - -dC / dln(f): derivative of C with ln(f), to identify inflection points.')
        print('Analysis functions:')
        print(' - Set area: can normalize input data. For proper data analysis the units should be [nF cm-2].')
        print('Further analysis:')
        print(' - Report inflection point for different T, then the fit activation energy.')
        print('   Traps can follow omega = 2 ksi T^2 exp(- E_omega / (kT)')
        print('   plot ln(omega T^-2) = ln(2 ksi) - E_omega / (kT)')
        return True



# -*- coding: utf-8 -*-
"""
Created on Tue Feb  6 19:49:25 2018

@author: Romain
"""

from os import path as ospath
import numpy as np


from grapa.graph import Graph
from grapa.graphIO import GraphIO
from grapa.datatypes.curveSpectrum import CurveSpectrum



class GraphSpectrum(Graph):
    """
    This is a generic class to open spectrum-like files. The purpose is to
    limit the endless multiplication of files in folder datatypes.
    Each child class of GraphSpectrum are constructed the same way as similar
    classes.
    Each types of 
    """
    
    FILEIO_GRAPHTYPE = 'Optical spectrum'

    AXISLABELS = [['Wavelength', '\lambda', 'nm'], ['Intensity', '', 'counts']]
    
    WHICHSUBCLASS = []
    
    @classmethod
    def isFileReadable(cls, fileName, fileExt, **kwargs):
        # ask every child class to know if it can open the file
        # if yes, remembers which one can
        GraphSpectrum.WHICHSUBCLASS = []
        for child in GraphSpectrum.__subclasses__():
            if child.isFileReadable(fileName, fileExt, **kwargs):
                GraphSpectrum.WHICHSUBCLASS.append(child)
        if len(GraphSpectrum.WHICHSUBCLASS) > 0:
            return True
        return False

    def readDataFromFile(self, attributes, **kwargs):
        # ask the child class which said they can, to open the file
        # By design readDataFromFile shall be called immediately after
        # isFileReadable, so the design is ~safe. Maybe we should explicitely 
        # perform the readable test again?
        for child in GraphSpectrum.WHICHSUBCLASS:
            res = child.readDataFromFile(self, attributes, **kwargs)
            if res is None or res != False:
                return True
        return False

        
            
            
class GraphSpectrumSpectraSuite(GraphSpectrum):
    """ Reads a HR2000 file (optical fiber spectrophotometer). """
    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', line2='', line3='', **kwargs):
        if fileExt == '.txt' and line1 == 'SpectraSuite Datei':
            return True
    def readDataFromFile(self, attributes, **kwargs):
        filenam_, fileext = ospath.splitext(self.filename) # , fileExt
        # TODO:  improve header parsing
        self.append(CurveSpectrum(np.transpose(np.genfromtxt(self.filename, skip_header=17, delimiter='\t', invalid_raise=False)), attributes))
        # BG = 1140 # seems not generally valid
        BG = 0
        self.curve(-1).setY(self.curve(-1).y() - BG)
        self.curve(-1).update({'label': filenam_.split('/')[-1]})
        self.headers.update({'collabels': ['Wavelength [nm]', 'Intensity [counts]']})
        self.update({'xlabel': self.formatAxisLabel(GraphSpectrum.AXISLABELS[0]),
                     'ylabel': self.formatAxisLabel(GraphSpectrum.AXISLABELS[1])})

        

class GraphSpectrumUVVIS(GraphSpectrum):
    """ Reads a ascii file from Shimadzu UVVIS 3600 """
    AXISLABELS = [['Wavelength', '', 'nm'], GraphSpectrum.AXISLABELS[1]]
    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', line2='', line3='', **kwargs):
        if (fileExt == '.txt'
            and line1.startswith('"') and line1.endswith('"')
            and line2.startswith('"Wavelength nm."	"') and line2.endswith('%"')):
            return True
    def readDataFromFile(self, attributes, **kwargs):
        GraphIO.readDataFromFileGeneric(self, attributes)
        sub = self.curve(-1).getAttribute('"Wavelength nm."')
        lbl = self.curve(-1).getAttribute('label')
        ylabel = GraphSpectrumUVVIS.AXISLABELS[1]
        if sub == 'R%':
            sub = 'CurveSpectrumReflectance'
            ylabel = ['Reflectance', '', '%']
            if lbl.endswith(' R R%'):
                lbl = lbl[:-3]+'%'
        elif sub == 'T%':
            sub = 'CurveSpectrumTransmittance'
            ylabel = ['Transmittance', '', '%']
            if lbl.endswith(' T T%'):
                lbl = lbl[:-3]+'%'
        else:
            sub = ''
        self.curve(-1).update({'_spectrumSubclass': sub, '_spectrumunit': '%'})
        self.castCurve('Curve Spectrum', self.length()-1, silentSuccess=True)
        self.curve(-1).update({'label': lbl, '"Wavelength nm."': ''})
        self.update({'xlabel': self.formatAxisLabel(GraphSpectrumUVVIS.AXISLABELS[0]),
                     'ylabel': self.formatAxisLabel(ylabel)})
        
        
    

        
class GraphSpectrumTRPLsetup(GraphSpectrum):
    """ TRPL setup PL spectrum """
    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', line2='', line3='', **kwargs):
        if fileExt == '.dat' and line1[:29] == 'Excitation Wavelength[nm]	crv':
            return True
        return False
    def readDataFromFile(self, attributes, **kwargs):
        len0 = self.length()
        GraphIO.readDataFromFileGeneric(self, attributes)
        self.castCurve('Curve Spectrum', len0, silentSuccess=True)
        self.curve(len0).update({'label': self.curve(len0).getAttribute('label').replace(' crv[0] [Cnts.]','')})
        self.update({'xlabel': self.formatAxisLabel(GraphSpectrum.AXISLABELS[0]),
                     'ylabel': self.formatAxisLabel(GraphSpectrum.AXISLABELS[1])})
        self.update({'subplots_adjust': [0.2, 0.15]})

        

class GraphSpectrumPerkinElmerASC(GraphSpectrum):
    """
    Opens a kind of Perkin Elmer file.
    This parser is not exhaustive and does not necessary comply with the
    file format specifications. Much more information is contained in the file
    """
    
    AXISLABELS = [GraphSpectrum.AXISLABELS[0], ['Intensity', '', '%']]
    
    @classmethod
    def isFileReadable(cls, fileName, fileExt, line1='', line2='', line3='', **kwargs):
        if fileExt == '.asc' and line3.endswith('.asc'):
            return True
        return False
    def readDataFromFile(self, attributes, **kwargs):
        # reads header and count headers lines
        nlines = 0
        headers = []
        with open(self.filename) as fp:
            for line in fp:
                line = line.strip()
                nlines += 1
                # parse a few possibly useful headers lines
                if len(line) > 0 and ' ' in line:
                    if line[0] not in ['0','1','2','3','4','5','6','7','8','9']:
                        headers.append(line)
                if line == '%R':
                    attributes.update({'_spectrumSubclass': 'CurveSpectrumReflectance'})
                if line == '%T':
                    attributes.update({'_spectrumSubclass': 'CurveSpectrumTransmittance'})
                if line == '#DATA':
                    break
        data = np.transpose(np.genfromtxt(self.filename, skip_header=nlines, 
                                          delimiter='\t', invalid_raise=False))
        self.append(CurveSpectrum(data, attributes))
        for h in range(len(headers)):
            self.curve(-1).update({'headers'+str(h): headers[h]})
        self.update({'xlabel': self.formatAxisLabel(GraphSpectrumPerkinElmerASC.AXISLABELS[0]),
                     'ylabel': self.formatAxisLabel(GraphSpectrumPerkinElmerASC.AXISLABELS[1])})
